# ruff: noqa: N806
from __future__ import annotations

import numpy as np
from f16 import GRAV_FTS2, SubsonicF16

import archimedes as arc
from archimedes import struct
from archimedes.spatial import EulerAngles, euler_kinematics

__all__ = ["trim", "TrimPoint", "TrimCondition", "TrimVariables"]


@struct
class TrimCondition:
    vt: float  # True airspeed [ft/s]
    alt: float = 0.0  # Altitude [ft]
    gamma: float = 0.0  # Flight path angle [deg]
    roll_rate: float = 0.0  # Roll rate [rad/s]
    pitch_rate: float = 0.0  # Pitch rate [rad/s]
    turn_rate: float = 0.0  # Turn rate [rad/s]


@struct
class TrimVariables:
    alpha: float  # Angle of attack [rad]
    beta: float  # Sideslip angle [rad]
    throttle: float  # Throttle setting [0-1]
    elevator: float  # Elevator deflection [deg]
    aileron: float  # Aileron deflection [deg]
    rudder: float  # Rudder deflection [deg]

    @property
    def inputs(self) -> SubsonicF16.Input:
        return SubsonicF16.Input(
            throttle=self.throttle,
            elevator=self.elevator,
            aileron=self.aileron,
            rudder=self.rudder,
        )


@struct
class TrimPoint:
    condition: TrimCondition
    variables: TrimVariables
    xcg: float  # CG location [-]
    name: str = ""
    description: str = ""
    state: SubsonicF16.State | None = None
    residuals: np.ndarray | None = None

    @property
    def inputs(self) -> SubsonicF16.Input:
        return self.variables.inputs


def trim_state(
    params: TrimVariables,
    condition: TrimCondition,
    model: SubsonicF16,
) -> SubsonicF16.State:
    gamma = np.deg2rad(condition.gamma)
    alpha = params.alpha
    beta = params.beta

    # Turn constraint (determines roll angle)
    G = condition.turn_rate * condition.vt / GRAV_FTS2  # Centripetal acceleration [g's]
    a = 1 - G * np.tan(alpha) * np.sin(beta)
    b = np.sin(gamma) / np.cos(beta)
    c = 1 + G**2 * np.cos(beta) ** 2

    num = (
        G
        * np.cos(beta)
        * np.sqrt(
            (a - b**2) + b * np.tan(alpha) * (c * (1 - b**2) + G**2 * np.sin(beta) ** 2)
        )
    )
    den = np.cos(alpha) * (a**2 - b**2 * (1 + c * np.tan(alpha) ** 2))
    phi = np.arctan2(num, den)

    # Rate-of-climb constraint (determines pitch angle)
    a = np.cos(alpha) * np.cos(beta)
    b = np.sin(phi) * np.sin(beta) + np.cos(phi) * np.sin(alpha) * np.cos(beta)
    num = a * b + np.sin(gamma) * np.sqrt(a**2 + b**2 - np.sin(gamma) ** 2)
    den = a**2 - np.sin(gamma) ** 2
    theta = np.arctan2(num, den)

    # Calculate the angular velocity based on the Euler rates and
    # roll-pitch-yaw angles (inverse Euler kinematics)
    rpy = np.hstack([phi, theta, 0.0])  # Arbitrary yaw angle
    H_inv = euler_kinematics(rpy, inverse=True)
    w_B = H_inv @ np.hstack(
        [condition.roll_rate, condition.pitch_rate, condition.turn_rate]
    )

    # Body-frame velocity (rotate from wind frame)
    v_W = np.array([condition.vt, 0.0, 0.0])  # Wind-frame velocity [ft/s]
    R_BW = EulerAngles([-beta, alpha], "zy").as_matrix()
    v_B = R_BW @ v_W

    att = EulerAngles(rpy)

    return model.State(
        pos=np.hstack([0.0, 0.0, -condition.alt]),
        att=att,
        v_B=v_B,
        w_B=w_B,
        eng=model.engine.trim(params.throttle),
        aero=model.aero.trim(),
        elevator=model.elevator.trim(params.elevator),
        aileron=model.aileron.trim(params.aileron),
        rudder=model.rudder.trim(params.rudder),
    )


def trim_residual(
    params: TrimVariables,
    condition: TrimCondition,
    model: SubsonicF16,
) -> np.ndarray:
    x = trim_state(params, condition, model)
    u = params.inputs
    x_t = model.dynamics(0.0, x, u)
    return np.hstack([x_t.v_B, x_t.w_B])


def trim(
    model: SubsonicF16,
    vt: float,  # True airspeed [ft/s]
    alt: float = 0.0,  # Altitude [ft]
    gamma: float = 0.0,  # Flight path angle [deg]
    roll_rate: float = 0.0,  # Roll rate [rad/s]
    pitch_rate: float = 0.0,  # Pitch rate [rad/s]
    turn_rate: float = 0.0,  # Turn rate [rad/s]
) -> TrimPoint:
    condition = TrimCondition(
        vt=vt,
        alt=alt,
        gamma=gamma,
        roll_rate=roll_rate,
        pitch_rate=pitch_rate,
        turn_rate=turn_rate,
    )
    params_guess = TrimVariables(
        alpha=0.0,
        beta=0.0,
        throttle=0.0,
        elevator=0.0,
        aileron=0.0,
        rudder=0.0,
    )

    params_guess_flat, unravel = arc.tree.ravel(params_guess)

    def residual(params_flat):
        params = unravel(params_flat)
        return trim_residual(params, condition, model)

    params_opt_flat = arc.root(residual, params_guess_flat)
    params_opt = unravel(params_opt_flat)

    state_opt = trim_state(params_opt, condition, model)

    return TrimPoint(
        condition=condition,
        variables=params_opt,
        state=state_opt,
        xcg=model.xcg,
        residuals=residual(params_opt_flat),
    )



# This code modifies code from Flax
#
# Copyright (c) 2024 The Flax Authors
# Licensed under Apache License 2.0
# https://github.com/google/flax
#
# Modifications and additions to the original code:
# Copyright (c) 2025 Pine Tree Labs, LLC
# Licensed under the GNU General Public License v3.0

# SPDX-FileCopyrightText: 2024 The Flax Authors
# SPDX-FileCopyrightText: 2025 Pine Tree Labs, LLC
# SPDX-License-Identifier: GPL-3.0-or-later
#
# As a combined work, use of this code requires compliance with the GNU GPL v3.0.
# The original license terms are included below for attribution:
#
# === Apache License 2.0 ===
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilities for defining custom classes that can be used with tree transformations.

This module provides tools for creating structured data types that work seamlessly
with Archimedes' tree functions. These tools are built on Python's dataclasses
with extensions for tree-specific behavior.

The module re-exports several names from the dataclasses module:

InitVar : Type annotation for init-only variables in dataclasses
    Used to mark fields that should be passed to __post_init__ but not stored.

fields : Function to retrieve fields of a dataclass
    Returns a list of Field objects representing the fields of the dataclass.
    This is useful for introspection and validation of dataclass instances.

replace : Function to create a new dataclass instance with updated fields
    For tree nodes created with @struct, use the .replace() method instead.
"""

from __future__ import annotations

import dataclasses
import functools
from collections.abc import Callable
from dataclasses import InitVar, fields, replace
from typing import Any, TypeVar

from typing_extensions import dataclass_transform

from ._registry import register_dataclass

__all__ = [
    "field",
    "struct",
    "InitVar",
    "is_struct",
    "fields",
    "replace",
]


T = TypeVar("T")


def field(
    static: bool = False,
    *,
    metadata: dict[str, Any] | None = None,
    **kwargs,
) -> dataclasses.Field:
    """
    Create a field specification with struct-related metadata.

    This function extends :py:func:`dataclasses.field()` with additional metadata to
    control how fields are treated in tree operations. Fields can be marked as static
    (metadata) or dynamic (data). Except for the `static` argument, all other arguments
    are passed directly to :py:func:`dataclasses.field()`; see documentation for the
    :py:mod:`dataclasses` module for details.

    Parameters
    ----------
    static : bool, default=False
        If True, the field is treated as static metadata rather than dynamic data.
        Static fields are preserved during tree transformations but not included
        in the flattened representation.
    metadata : dict, optional
        Additional metadata to include in the field specification. This will be
        merged with the ``static`` setting.
    **kwargs : dict
        Additional keyword arguments passed to :py:func:`dataclasses.field()`.

    Returns
    -------
    field_object : dataclasses.Field
        A field specification with the appropriate metadata.

    Notes
    -----
    When to use:

    - To mark configuration parameters that shouldn't change during operations
    - To define default values or constructors for fields

    Static fields are not included when you flatten tree-structured data or apply
    transformations like ``map``, but they are preserved in the structure and included
    when you reconstruct the object.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> @arc.struct
    >>> class Vehicle:
    ...     # Dynamic state variables (included in flattening)
    ...     position: np.ndarray
    ...     velocity: np.ndarray
    ...
    ...     # Static configuration parameters (excluded from flattening)
    ...     mass: float = arc.field(static=True, default=1000.0)
    ...     drag_coef: float = arc.field(static=True, default=0.3)
    ...
    ...     # With additional metadata
    ...     name: str = arc.field(
    ...         static=True,
    ...         default="vehicle",
    ...         metadata={"description": "Vehicle identifier"}
    ...     )
    >>>
    >>> # Create an instance
    >>> car = Vehicle(
    ...     position=np.array([0.0, 0.0]),
    ...     velocity=np.array([10.0, 0.0]),
    ... )
    >>>
    >>> # When flattened, only dynamic fields are included
    >>> flat, _ = arc.tree.flatten(car)
    >>> print(len(flat))  # Only position and velocity are included
    2

    See Also
    --------
    struct : Decorator for creating tree-compatible dataclasses
    register_dataclass : Register a dataclass as compatible with tree operations
    """
    f: dataclasses.Field = dataclasses.field(
        metadata=(metadata or {}) | {"static": static},
        **kwargs,
    )
    return f


@dataclass_transform(field_specifiers=(field,))  # type: ignore[literal-required]
def struct(cls: T | None = None, **kwargs) -> T | Callable:
    """
    Decorator to convert a class into a tree-compatible frozen dataclass.

    This decorator creates a structured data class that can be seamlessly used
    with Archimedes' tree functions. The class will be registered with the tree
    system, allowing its instances to be flattened, mapped over, and transformed
    while preserving its structure.

    Parameters
    ----------
    cls : type, optional
        The class to convert into a tree-compatible dataclass.
    **kwargs : dict
        Additional keyword arguments passed to dataclasses.dataclass().
        By default, ``frozen=True`` is set unless explicitly overridden.

    Returns
    -------
    decorated_class : type
        The decorated class, now a frozen dataclass registered as tree-compatible.

    Notes
    -----

    The "frozen" attribute makes the class immutable, meaning that once an instance
    is created, its fields cannot be modified. This is useful for ensuring that
    the state of the object remains consistent during operations. The ``replace()``
    method allows you to create modified copies of the object with new values for
    specific fields.

    Fields are automatically classified as either "data" (dynamic values that
    change during operations) or "static" (configuration parameters). By default,
    all fields are treated as data unless marked with ``field(static=True)``.

    The decorated class:

    - Is frozen (immutable) by default
    - Has a ``replace()`` method for creating modified copies
    - Will be properly handled by ``tree.flatten()``, ``tree.map()``, etc.
    - Can be nested within other tree nodes (structs, dicts, tuples, etc.)

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> @arc.struct
    >>> class Vehicle:
    ...     # Dynamic state variables (included in transformations)
    ...     position: np.ndarray
    ...     velocity: np.ndarray
    ...
    ...     # Static configuration parameters (preserved during transformations)
    ...     mass: float = arc.field(static=True, default=1000.0)
    ...     drag_coef: float = arc.field(static=True, default=0.3)
    ...
    ...     def kinetic_energy(self):
    ...         return 0.5 * self.mass * np.sum(self.velocity**2)
    >>>
    >>> # Create an instance
    >>> car = Vehicle(
    ...     position=np.array([0.0, 0.0]),
    ...     velocity=np.array([10.0, 0.0]),
    ... )
    >>>
    >>> # Create a modified copy
    >>> car2 = car.replace(position=np.array([5.0, 0.0]))
    >>>
    >>> # Apply a transformation (only to dynamic fields)
    >>> scaled = arc.tree.map(lambda x: x * 2, car)
    >>> print(scaled.position)    # [0. 0.] -> [0. 0.]
    >>> print(scaled.velocity)    # [10. 0.] -> [20. 0.]
    >>> print(scaled.mass)        # 1000.0 (unchanged)
    >>>
    >>> # Nested structs
    >>> @arc.struct
    >>> class System:
    ...     vehicle1: Vehicle
    ...     vehicle2: Vehicle
    ...
    ...     def total_energy(self):
    ...         return self.vehicle1.kinetic_energy() + self.vehicle2.kinetic_energy()
    >>>
    >>> system = System(car, car2)
    >>> # This transformation applies to all dynamic fields in the entire hierarchy
    >>> scaled_system = arc.tree.map(lambda x: x * 0.5, system)

    See Also
    --------
    field : Define fields with tree-specific metadata
    """
    # Support passing arguments to the decorator (e.g. @struct(kw_only=True))
    if cls is None:
        return functools.partial(struct, **kwargs)

    # check if already recognized as a tree node
    if "_arc_struct" in cls.__dict__:
        return cls

    if "frozen" not in kwargs.keys():
        kwargs["frozen"] = True
    data_cls = dataclasses.dataclass(**kwargs)(cls)  # type: ignore
    meta_fields = []
    data_fields = []
    for field_info in dataclasses.fields(data_cls):
        # if not field_info.init:
        #     continue
        is_static = field_info.metadata.get("static", False)
        if not is_static:
            data_fields.append(field_info.name)
        else:
            meta_fields.append(field_info.name)

    def replace(self, **updates) -> T:
        """Returns a new object replacing the specified fields with new values."""
        new: T = dataclasses.replace(self, **updates)
        return new

    data_cls.replace = replace

    register_dataclass(data_cls, data_fields, meta_fields)

    # add a _arc_struct flag to distinguish from regular dataclasses
    data_cls._arc_struct = True  # type: ignore[attr-defined]

    return data_cls  # type: ignore


def is_struct(obj: Any) -> bool:
    """
    Check if an object is a registered struct class.

    This function determines whether an object was created using the
    :py:func:`struct` decorator, which indicates it has special handling
    for tree operations.

    Parameters
    ----------
    obj : Any
        The object to check.

    Returns
    -------
    is_node : bool
        ``True`` if the object is a struct created with the decorator,
        ``False`` otherwise.

    Notes
    -----
    When to use:

    - To check if an object will be handled specially by tree operations
    - For conditional logic based on whether an object is a custom struct
    - For debugging tree-related functionality

    This function specifically checks for objects created with the
    :py:func:`struct` decorator, not built-in structured data types like lists,
    tuples, and dictionaries.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> @arc.struct
    >>> class State:
    ...     x: np.ndarray
    ...     v: np.ndarray
    >>>
    >>> state = State(np.zeros(3), np.ones(3))
    >>> print(arc.tree.is_struct(state))
    True
    >>>
    >>> # Regular dataclass is not a struct
    >>> from dataclasses import dataclass
    >>>
    >>> @dataclass
    >>> class RegularState:
    ...     x: np.ndarray
    ...     v: np.ndarray
    >>>
    >>> regular_state = RegularState(np.zeros(3), np.ones(3))
    >>> print(arc.tree.is_struct(regular_state))
    False
    >>>
    >>> # Built-in containers aren't custom structs
    >>> print(arc.tree.is_struct({"x": np.zeros(3)}))
    False

    See Also
    --------
    struct : Decorator for creating tree-compatible dataclasses
    """
    return hasattr(obj, "_arc_struct")



import numpy as np

__all__ = ["wind_frame"]


def wind_frame(v_rel_B):
    """Compute total velocity, angle of attack, and sideslip angle

    The input should be the vehicle wind-relative velocity computed in
    body-frame axes.  If the inertial velocity of the vehicle expressed in
    body-frame axes is v_B and the Earth-relative wind velocity is w_N,
    then the relative velocity is v_rel_B = v_B + R_BN @ w_N, where R_BN
    is the rotation matrix from inertial to body frame.

    If there is no wind, then v_rel_B = v_B.
    """
    u, v, w = v_rel_B
    vt = np.sqrt(u**2 + v**2 + w**2)
    alpha = np.arctan2(w, u)
    beta = np.arcsin(v / vt)
    return vt, alpha, beta



# This code combines and modifies code from multiple sources:

# 1. Originally derived from pybaum
#    Original Copyright (c) 2022 Janoś Gabler, Tobias Raabe
#    Licensed under MIT
#    https://github.com/OpenSourceEconomics/pybaum

# 2. Incorporates code from JAX
#    Copyright (c) 2021 The JAX Authors
#    Licensed under Apache License 2.0
#    https://github.com/jax-ml/jax

# Modifications and additions to the original code:
# Copyright (c) 2025 Pine Tree Labs, LLC
# Licensed under the GNU General Public License v3.0

# SPDX-FileCopyrightText: 2021 The JAX Authors
# SPDX-FileCopyrightText: 2022 Janoś Gabler, Tobias Raabe
# SPDX-FileCopyrightText: 2025 Pine Tree Labs, LLC
# SPDX-License-Identifier: GPL-3.0-or-later

# As a combined work, use of this code requires compliance with the GNU GPL v3.0.
# The original license terms are included below for attribution:

# === MIT License ===
# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the "Software"), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to the following
# conditions:

# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# === Apache License 2.0 ===
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from collections import OrderedDict, namedtuple
from functools import partial
from typing import NamedTuple

import numpy as np
import pytest

from archimedes import StructConfig, UnionConfig, field, struct, tree
from archimedes._core import SymbolicArray, compile, sym
from archimedes.tree._tree_util import NONE_DEF


def tree_equal(a, b):
    return tree.all(tree.map(lambda a, b: np.allclose(a, b), a, b))


def assert_flat_equal(f1, f2):
    assert len(f1) == len(f2)
    for a, b in zip(f1, f2):
        if isinstance(a, np.ndarray):
            assert np.allclose(a, b)
        else:
            assert a == b


def _example_tree():
    return (
        [0, np.array([1, 2]), {"a": (3.0, 4), "b": 5}],
        6,
    )


@pytest.fixture
def example_tree():
    return _example_tree()


@pytest.fixture
def example_flat():
    return [0, np.array([1, 2]), 3.0, 4, 5, 6]


@pytest.fixture
def example_treedef():
    return tree.structure(_example_tree())


def test_flatten(example_tree, example_flat):
    flat, treedef = tree.flatten(example_tree)

    assert_flat_equal(flat, example_flat)

    unflat = treedef.unflatten(flat)
    assert unflat == example_tree


def test_flatten_with_is_leaf(example_tree):
    flat, _ = tree.flatten(
        example_tree,
        is_leaf=lambda tree: isinstance(tree, dict),
    )
    expected_flat = [0, np.array([1, 2]), {"a": (3.0, 4), "b": 5}, 6]
    assert_flat_equal(flat, expected_flat)


def test_unflatten(example_flat, example_treedef, example_tree):
    unflat = tree.unflatten(example_treedef, example_flat)
    assert tree_equal(unflat, example_tree)


def test_unflatten_with_is_leaf(example_tree):
    treedef = tree.structure(example_tree, is_leaf=lambda tree: isinstance(tree, dict))
    flat = tree.leaves(example_tree, is_leaf=lambda tree: isinstance(tree, dict))
    unflat = treedef.unflatten(flat)
    assert tree_equal(unflat, example_tree)


def test_map():
    t = [{"a": 1, "b": 2, "c": {"d": 3, "e": 4}}]
    calculated = tree.map(lambda x: x * 2, t)
    expected = [{"a": 2, "b": 4, "c": {"d": 6, "e": 8}}]
    assert calculated == expected


def test_multimap():
    t = [{"a": 1, "b": 2, "c": {"d": 3, "e": 4}}]
    mapped = tree.map(lambda x: x**2, t)
    multimapped = tree.map(lambda x, y: x * y, t, t)
    assert mapped == multimapped

    # Incorrect structure
    with pytest.raises(ValueError, match=r"Trees must have the same structure.*"):
        tree.map(lambda x, y: x * y, t, (0, 2, 3))


def test_reduce():
    t = [{"a": 1, "b": 2, "c": {"d": 3, "e": 4}}]
    _sum = tree.reduce(lambda x, y: x + y, t, 0)
    assert _sum == sum(range(1, 5))


def test_flatten_none():
    flat, treedef = tree.flatten(None)
    assert flat == []
    assert treedef is NONE_DEF


def test_flatten_namedtuple():
    bla = namedtuple("bla", ["a", "b"])(1, 2)
    flat, _ = tree.flatten(bla)
    assert flat == [1, 2]


def test_namedtuple_class():
    class Point(NamedTuple):
        x: float
        y: float

    p = Point(1.0, 2.0)
    flat, treedef = tree.flatten(p)
    assert flat == [1.0, 2.0]

    unflat = treedef.unflatten(flat)
    assert unflat == p


def test_ordered_dict():
    d = OrderedDict({"a": 1, "b": 2})
    flat, treedef = tree.flatten(d)
    assert flat == [1, 2]

    unflat = treedef.unflatten(flat)
    assert unflat == d


def test_custom_node():
    class Point:
        def __init__(self, x, y, name):
            self.x = x
            self.y = y
            self.name = name

    def point_flatten(point: Point):
        aux_data = (point.name,)
        children = (point.x, point.y)
        return children, aux_data

    def point_unflatten(aux_data, children) -> Point:
        x, y = children
        (name,) = aux_data
        return Point(x, y, name)

    tree.register_struct(Point, point_flatten, point_unflatten)

    p = Point(1.0, 2.0, "p")
    flat, treedef = tree.flatten(p)
    assert flat == [1.0, 2.0]

    unflat = treedef.unflatten(flat)
    assert unflat.x == p.x
    assert unflat.y == p.y
    assert unflat.name == p.name


def test_register_dataclass_field():
    # Mark static fields using `field` metadata
    @tree.register_dataclass
    @dataclasses.dataclass
    class Point:
        x: float
        y: float
        name: str = dataclasses.field(metadata=dict(static=True))

    p = Point(1.0, 2.0, "p")
    flat, treedef = tree.flatten(p)
    assert flat == [1.0, 2.0]

    unflat = treedef.unflatten(flat)
    assert unflat.x == p.x
    assert unflat.y == p.y
    assert unflat.name == p.name


def test_register_dataclass_metadata():
    # Manually mark fields as static or data
    @partial(
        tree.register_dataclass,
        meta_fields=("name",),
        data_fields=("x", "y"),
    )
    @dataclasses.dataclass
    class Point:
        x: float
        y: float
        name: str

    p = Point(1.0, 2.0, "p")
    flat, treedef = tree.flatten(p)
    assert flat == [1.0, 2.0]

    unflat = treedef.unflatten(flat)
    assert unflat.x == p.x
    assert unflat.y == p.y
    assert unflat.name == p.name

    # Error handling
    with pytest.raises(TypeError, match=r".*data_fields and meta_fields must both.*"):

        @partial(
            tree.register_dataclass,
            meta_fields=("name",),
        )
        @dataclasses.dataclass
        class Point:
            x: float
            y: float
            name: str

    with pytest.raises(TypeError, match=r".*fields are required.*"):

        @tree.register_dataclass
        class Point:
            x: float
            y: float

    with pytest.raises(ValueError, match=r".*``init=True`` and only them."):

        @partial(
            tree.register_dataclass,
            meta_fields=("name",),
            data_fields=("x", "y"),
        )
        @dataclasses.dataclass
        class Point:
            x: float
            y: float = dataclasses.field(init=False)

    with pytest.raises(ValueError, match=r".*Missing fields.*"):

        @partial(
            tree.register_dataclass,
            meta_fields=("name",),
            data_fields=("x"),
        )
        @dataclasses.dataclass
        class Point:
            x: float
            y: float


class TestRavel:
    def _test_ravel_unravel(self, struct, compile_fn=False):
        flat, unravel = tree.ravel(struct)
        assert isinstance(flat, SymbolicArray)

        if compile_fn:
            unravel = compile(unravel)

        # Check that the unraveling works
        unraveled = unravel(flat)

        for x, y in zip(struct, unraveled):
            assert x.shape == y.shape
            assert x.dtype == y.dtype

        return unraveled

    def test_ravel_none(self):
        flat, unravel = tree.ravel(None)
        assert flat.size == 0
        assert unravel(flat) is None

    def test_ravel_identity(self):
        x = sym("x", shape=(3,), dtype=np.float64)
        flat, unravel = tree.ravel(x)
        assert isinstance(flat, SymbolicArray)
        unraveled = unravel(flat)
        assert isinstance(unraveled, SymbolicArray)

    def test_ravel_scalar(self):
        x = 0.0
        flat, unravel = tree.ravel(x)
        assert flat == 0
        unraveled = unravel(flat)
        assert unraveled == 0

    def test_ravel_tuple(self):
        struct = []
        for i, (shape, dtype) in enumerate(
            [
                ((2, 3), np.float64),
                ((3,), np.int32),
                ((2, 5), np.int32),
                ((), np.bool_),
            ]
        ):
            struct.append(sym(f"x{i}", shape=shape, dtype=dtype))
        struct = tuple(struct)
        self._test_ravel_unravel(struct)

    def test_ravel_list(self):
        struct = []
        for i, (shape, dtype) in enumerate(
            [
                ((2, 3), np.float64),
                ((3,), np.int32),
                ((2, 5), np.int32),
                ((), np.bool_),
            ]
        ):
            struct.append(sym(f"x{i}", shape=shape, dtype=dtype))
        self._test_ravel_unravel(struct)

    def test_ravel_namedtuple(self):
        class Point(NamedTuple):
            x: SymbolicArray
            y: SymbolicArray
            z: SymbolicArray

        struct = Point(
            sym("x", shape=(3,), dtype=np.float64),
            sym("y", shape=(3,), dtype=np.float64),
            sym("z", shape=(3,), dtype=np.float64),
        )

        unraveled = self._test_ravel_unravel(struct)
        assert isinstance(unraveled, Point)

        # Test that this still works when the unravel function is compiled
        # See https://github.com/PineTreeLabs/archimedes/issues/30 for previous bug
        unraveled = self._test_ravel_unravel(struct, compile_fn=True)
        assert isinstance(unraveled, Point)

    def test_ravel_dict(self):
        struct = {
            "a": sym("a", shape=(2, 3), dtype=np.float64),
            "b": sym("b", shape=(3,), dtype=np.int32),
            "c": sym("c", shape=(2, 5), dtype=np.int32),
            "d": sym("d", shape=(), dtype=np.bool_),
        }

        flat, unravel = tree.ravel(struct)
        assert isinstance(flat, SymbolicArray)

        # Check that the unraveling works
        unraveled = unravel(flat)

        for x, y in zip(struct.values(), unraveled.values()):
            assert x.shape == y.shape
            assert x.dtype == y.dtype

    def test_ravel_error_handling(self):
        struct = "abc"
        with pytest.raises(TypeError):
            tree.ravel(struct)

    def test_ravel_nested(self):
        struct = (4.0, 3.0), np.array([2, 1])
        flat, unravel = tree.ravel(struct)

        assert np.allclose(flat, [4, 3, 2, 1])
        unflat = unravel(flat)
        assert tree_equal(struct, unflat)

    def test_ravel_hash_dict(self):
        # Test that the keys of a dictionary are hashed, so
        # changing the keys will force recompilation.

        @compile
        def f(x, p):
            a = p.get("a", 0)
            b = p.get("b", 0)
            return a * x[0] + b * x[1]

        x = np.array([1, 2])

        assert f(x, {"a": 1}) == x[0]
        assert f(x, {"b": 1}) == x[1]


def test_register_struct():
    # Mark static fields using `field`
    @struct(kw_only=True)
    class Point:
        x: float
        y: float
        name: str = field(static=True)

    assert tree.is_struct(Point)

    p = Point(x=1.0, y=2.0, name="p")

    # Check that the dataclass is frozen
    with pytest.raises(dataclasses.FrozenInstanceError):
        p.x = 3.0

    flat, treedef = tree.flatten(p)
    assert flat == [1.0, 2.0]

    unflat = treedef.unflatten(flat)
    assert unflat.x == p.x
    assert unflat.y == p.y
    assert unflat.name == p.name

    p2 = p.replace(y=0.0)
    assert p2 is not p
    assert p2.y == 0.0
    assert p.y == 2.0

    # Check that re-applying the struct does nothing
    # because of the `_arc_struct` class attribute.
    assert struct(Point) is Point


@struct
class VariantBase:
    pass


@struct(kw_only=True)  # Add keyword arg to test passing with cls=None
class VariantA(VariantBase):
    param: float  # m/s^2

    def __call__(self):
        return self.param


@struct(kw_only=True)
class VariantB(VariantBase):
    param: float  # m/s^2

    def __call__(self):
        return -self.param


# Test that we can create a base class with a shared config
class VariantConfigBase(StructConfig):
    param: float


class VariantAConfig(VariantConfigBase, type="positive"):
    def build(self):
        return VariantA(param=self.param)


class VariantBConfig(VariantConfigBase, type="negative"):
    def build(self):
        return VariantB(param=self.param)


def test_variant_config():
    UnionConfig[VariantAConfig, VariantBConfig]

    with pytest.raises(TypeError, match=r".*must be a subclass of StructConfig.*"):
        UnionConfig[VariantA, VariantB]

    # Single-variant config
    UnionConfig[VariantAConfig]


def test_struct_config():
    param = 42
    config = VariantAConfig(param=param)

    model = config.build()
    y = model()
    assert np.allclose(y, param)



import numpy as np
import pytest

from archimedes.experimental.coco.ocp import (
    BoundaryData,
    Constraint,
    MultiStageOptimalControlProblem,
    OptimalControlProblem,
    control_bounds,
    end_time,
    final_condition,
    initial_condition,
    parameter_bounds,
    start_time,
    state_bounds,
)


def test_boundary_data():
    t0, x0 = 0.0, np.array([1.0, 2.0])
    tf, xf = 1.0, np.array([3.0, 4.0])
    p = np.array([5.0])
    data = BoundaryData(t0, x0, tf, xf, p)
    assert data.t0 == t0
    assert np.array_equal(data.x0, x0)
    assert data.tf == tf
    assert np.array_equal(data.xf, xf)
    assert np.array_equal(data.p, p)


def test_constraint():
    def func(x):
        return x**2

    nc = 1
    c = Constraint(func, nc, lower_bound=np.array([0.0]), upper_bound=np.array([1.0]))
    assert c.func(2.0) == 4.0
    assert c.nc == nc
    assert np.array_equal(c.lower_bound, [0.0])
    assert np.array_equal(c.upper_bound, [1.0])


def test_initial_condition():
    x0 = np.array([1.0, 2.0])
    c = initial_condition(x0)
    data = BoundaryData(
        0.0, np.array([1.0, 2.0]), 1.0, np.array([3.0, 4.0]), np.array([])
    )
    assert np.array_equal(c(data), np.zeros(2))


def test_final_condition():
    xf = np.array([3.0, 4.0])
    c = final_condition(xf)
    data = BoundaryData(
        0.0, np.array([1.0, 2.0]), 1.0, np.array([3.0, 4.0]), np.array([])
    )
    assert np.array_equal(c(data), np.zeros(2))


def test_start_time():
    t0 = 0.0
    c = start_time(t0)
    data = BoundaryData(
        0.0, np.array([1.0, 2.0]), 1.0, np.array([3.0, 4.0]), np.array([])
    )
    assert c(data) == 0.0


def test_end_time():
    tf = 1.0
    c = end_time(tf)
    data = BoundaryData(
        0.0, np.array([1.0, 2.0]), 1.0, np.array([3.0, 4.0]), np.array([])
    )
    assert c(data) == 0.0


def test_parameter_bounds():
    lb, ub = np.array([-1.0]), np.array([1.0])
    c = parameter_bounds(lb, ub)
    p = np.array([0.5])
    data = BoundaryData(0.0, np.array([1.0, 2.0]), 1.0, np.array([3.0, 4.0]), p)
    assert np.array_equal(c(data), p)
    assert np.array_equal(c.lower_bound, lb)
    assert np.array_equal(c.upper_bound, ub)

    # Error case: bound dimension mismatch
    lb, ub = np.array([-1.0, -2.0]), np.array([1.0])
    with pytest.raises(ValueError):
        parameter_bounds(lb, ub)


def test_state_bounds():
    lb, ub = np.array([-1.0, -2.0]), np.array([1.0, 2.0])
    c = state_bounds(lb, ub)
    t, x, u, p = 0.0, np.array([0.5, 1.5]), np.array([0.0]), np.array([])
    assert np.array_equal(c(t, x, u, p), x)
    assert np.array_equal(c.lower_bound, lb)
    assert np.array_equal(c.upper_bound, ub)

    # Error case: bound dimension mismatch
    lb, ub = np.array([-1.0, -2.0]), np.array([1.0])
    with pytest.raises(ValueError):
        state_bounds(lb, ub)


def test_control_bounds():
    lb, ub = np.array([-1.0]), np.array([1.0])
    c = control_bounds(lb, ub)
    t, x, u, p = 0.0, np.array([0.0, 0.0]), np.array([0.5]), np.array([])
    assert np.array_equal(c(t, x, u, p), u)
    assert np.array_equal(c.lower_bound, lb)
    assert np.array_equal(c.upper_bound, ub)

    # Error case: bound dimension mismatch
    lb, ub = np.array([-1.0, -2.0]), np.array([1.0])
    with pytest.raises(ValueError):
        control_bounds(lb, ub)


def test_optimal_control_problem():
    nx, nu = 2, 1

    def ode(t, x, u, p):
        return np.array([x[1], u[0]], like=x)

    def quad(t, x, u, p):
        return x[0] ** 2 + u[0] ** 2

    def cost(x0, t0, xf, tf, q, p):
        return q

    boundary_constraints = [initial_condition(np.zeros(2)), final_condition(np.ones(2))]
    path_constraints = [state_bounds(np.array([-1.0, -1.0]), np.array([1.0, 1.0]))]

    ocp = OptimalControlProblem(
        nx,
        nu,
        ode=ode,
        quad=quad,
        cost=cost,
        boundary_constraints=boundary_constraints,
        path_constraints=path_constraints,
    )

    assert ocp.nx == nx
    assert ocp.nu == nu
    assert ocp.np == 0
    p = np.array([])
    assert np.allclose(
        ocp.ode(0.0, np.array([1.0, 2.0]), np.array([3.0]), p), np.array([2.0, 3.0])
    )
    assert ocp.quad(0.0, np.array([1.0, 2.0]), np.array([3.0]), p) == 10.0
    assert ocp.cost(np.array([1.0, 2.0]), 0.0, np.array([3.0, 4.0]), 1.0, 5.0, p) == 5.0
    assert len(ocp.boundary_constraints) == 2
    assert len(ocp.path_constraints) == 1


def test_multistage():
    # Construct a multistage problem
    nx, nu = 2, 1

    def ode(t, x, u, p):
        return np.array([x[1], u[0]], like=x)

    def quad(t, x, u, p):
        return x[0] ** 2 + u[0] ** 2

    def cost(x0, t0, xf, tf, q, p):
        return q

    stage1 = OptimalControlProblem(
        nx,
        nu,
        ode=ode,
        quad=quad,
        cost=cost,
    )

    stage2 = OptimalControlProblem(
        nx,
        nu,
        ode=ode,
        quad=quad,
        cost=cost,
    )

    def stage_constraint(bd):
        return bd[0].xf - bd[1].x0  # State continuity

    ocp = MultiStageOptimalControlProblem(
        stages=[stage1, stage2],
        stage_constraints=[Constraint(stage_constraint, nx)],
    )

    # TODO: Remove when implemented
    with pytest.raises(NotImplementedError):
        ocp.dynamics_residual(None, None, None, None, None)



# ruff: noqa: N802, N803, N806

import numpy as np
import pytest

import archimedes as arc
from archimedes.experimental import coco as cc
from archimedes.experimental.coco.interpolation import (
    LagrangePolynomial,
    gauss_legendre,
    gauss_lobatto,
    gauss_radau,
)


def test_gauss_legendre():
    x, w = gauss_legendre(5)
    assert len(x) == 5
    assert len(w) == 5
    assert np.isclose(np.sum(w), 2.0)


def test_gauss_radau():
    with pytest.raises(ValueError):
        gauss_radau(0)

    x, w = gauss_radau(5)
    assert len(x) == 5
    assert len(w) == 5
    assert x[0] == -1.0
    assert np.isclose(np.sum(w), 2.0)


def test_gauss_lobatto():
    with pytest.raises(ValueError):
        gauss_lobatto(1)

    x, w = gauss_lobatto(5)
    assert len(x) == 5
    assert len(w) == 5
    assert x[0] == -1.0
    assert x[-1] == 1.0
    assert np.isclose(np.sum(w), 2.0)


def test_lagrange_polynomial_interpolate():
    nodes = np.array([-1.0, 0.0, 1.0])
    poly = LagrangePolynomial(nodes)

    y0 = np.array([1.0, 0.0, 1.0])
    x = np.linspace(-1, 1, 11)
    y = arc.compile(poly.interpolate)(y0, x)

    assert np.isclose(y[0], 1.0)
    assert np.isclose(y[5], 0.0)
    assert np.isclose(y[-1], 1.0)


def test_lagrange_polynomial_diff_matrix():
    nodes = np.array([-1.0, 0.0, 1.0])
    poly = LagrangePolynomial(nodes)

    D = poly.diff_matrix
    assert D.shape == (3, 3)
    assert np.isclose(np.sum(D, axis=1), 0.0).all()


# Integrated tests for working with the LagrangePolynomial class
@pytest.mark.parametrize("gauss", [gauss_legendre, gauss_radau, gauss_lobatto])
def test_barycentric(gauss):
    N = 16
    nodes, weights = gauss(N)  # Gaussian quadrature weights

    assert len(nodes) == N
    assert len(weights) == N

    # If the endpoints are not included (i.e. Radau or Legendre),
    # add them for the differentiation matrix.  These points are
    # unweighted because they are not part of the quadrature rule.
    if nodes[0] != -1:  # Legendre only
        nodes = np.insert(nodes, 0, -1.0)
        weights = np.insert(weights, 0, 0.0)
    if nodes[-1] != 1:  # Both Radau and Legendre
        nodes = np.append(nodes, 1)
        weights = np.append(weights, 0.0)

    p = cc.LagrangePolynomial(nodes)

    # Shift [-1, 1] -> [a, b]
    a, b = -2, 5  # Interval
    x = (b - a) / 2 * nodes + (a + b) / 2

    D = (2 / (b - a)) * p.diff_matrix
    w = (b - a) / 2 * weights  # weights on shifted interval

    y = np.cos(x)  # Evaluate the function at the nodes

    # Test differentiation
    assert np.allclose(D @ y, -np.sin(x), atol=1e-8)

    # Test quadrature (more accurate than differentiation)
    I1 = np.dot(w, y)
    I2 = np.sin(b) - np.sin(a)

    assert np.allclose(I1, I2, atol=1e-12)


def test_edge_cases():
    with pytest.raises(ValueError):
        cc.gauss_radau(0)

    cc.gauss_radau(1)
    cc.gauss_radau(2)

    with pytest.raises(ValueError):
        cc.gauss_lobatto(0)
        cc.gauss_lobatto(1)

    cc.gauss_lobatto(2)
    cc.gauss_lobatto(3)


def test_interpolation():
    N = 16
    nodes, _weights = cc.gauss_legendre(N)
    p = cc.LagrangePolynomial(nodes)

    f = np.cos
    y0 = f(nodes)

    x = 0.5 * (nodes[5] + nodes[6])
    y = f(x)
    y_interp = arc.compile(p.interpolate)(y0, x)

    assert np.allclose(y, y_interp, atol=1e-8)

    # Incorrect number of nodes in call to interpolate
    with pytest.raises(ValueError):
        p.interpolate(y0[:-1], x)



from __future__ import annotations

import abc

import numpy as np

from archimedes import StructConfig, UnionConfig, field, struct

__all__ = [
    "Actuator",
    "IdealActuator",
    "IdealActuatorConfig",
    "LagActuator",
    "LagActuatorConfig",
    "ActuatorConfig",
]


class Actuator(metaclass=abc.ABCMeta):
    """Abstract base class for SISO actuator components."""

    @struct
    class State:
        pass  # No state by default

    def dynamics(self, t: float, x: State, u: float) -> State:
        """Compute the actuator state derivative."""
        return x

    @abc.abstractmethod
    def output(self, t: float, x: State, u: float) -> float:
        """Compute the actuator output."""
        pass

    def trim(self, command: float) -> State:
        """Return a steady-state actuator state for the given command."""
        return self.State()


@struct
class IdealActuator(Actuator):
    """Ideal actuator with instantaneous linear response."""

    gain: float = 1.0

    def output(self, t: float, x: Actuator.State, u: float) -> float:
        return self.gain * u


class IdealActuatorConfig(StructConfig, type="ideal"):
    gain: float = 1.0

    def build(self) -> IdealActuator:
        return IdealActuator(self.gain)


@struct
class LagActuator(Actuator):
    tau: float
    gain: float = 1.0
    rate_limit: float | None = field(static=True, default=None)
    pos_limits: tuple[float, float] | None = field(static=True, default=None)

    @struct
    class State(Actuator.State):
        position: float = 0.0

    def dynamics(self, t: float, x: State, u: float) -> State:
        # Compute desired rate
        cmd, pos = u, x.position
        rate = (self.gain * cmd - pos) / self.tau

        # Apply rate limit
        if self.rate_limit is not None:
            max_rate = self.rate_limit
            rate = np.clip(rate, -max_rate, max_rate)

        if self.pos_limits is not None:
            min_pos, max_pos = self.pos_limits
            rate = np.where((pos <= min_pos) * (rate < 0.0), 0.0, rate)
            rate = np.where((pos >= max_pos) * (rate > 0.0), 0.0, rate)

        return self.State(rate)

    def output(self, t: float, x: State, u: float) -> float:
        pos = x.position
        if self.pos_limits is not None:
            min_pos, max_pos = self.pos_limits
            pos = np.clip(pos, min_pos, max_pos)
        return pos

    def trim(self, command: float) -> State:
        pos = command
        if self.pos_limits is not None:
            min_pos, max_pos = self.pos_limits
            pos = np.clip(pos, min_pos, max_pos)
        return self.State(position=pos)


class LagActuatorConfig(StructConfig, type="lag"):
    tau: float  # Time constant [sec]
    gain: float = field(default=1.0)
    rate_limit: float | None = field(default=None)  # Rate limit [units/sec]
    pos_limits: tuple[float, float] | None = field(
        default=None
    )  # Position limits [units]

    def build(self) -> LagActuator:
        return LagActuator(
            tau=self.tau,
            gain=self.gain,
            rate_limit=self.rate_limit,
            pos_limits=self.pos_limits,
        )


ActuatorConfig = UnionConfig[
    IdealActuatorConfig,
    LagActuatorConfig,
]



import numpy as np

from archimedes import tree

__all__ = ["Timeseries"]


@tree.struct
class Timeseries:
    """Container for synchronized input-output time series data.

    This class provides a structured way to organize and validate time series
    data for system identification applications. It ensures that input signals,
    output measurements, and time vectors are consistently sized and properly
    formatted for use with Kalman filters and parameter estimation algorithms.

    The class is implemented as a `@struct`, making it compatible with tree
    operations throughout the Archimedes framework.

    Parameters
    ----------
    ts : array_like
        Time vector of shape ``(N,)`` containing monotonic time samples.
        Must be one-dimensional.
    us : array_like
        Input signal matrix of shape ``(nu, N)`` where ``nu`` is the number
        of input channels and ``N`` is the number of time samples. Each row
        represents one input channel over time.
    ys : array_like
        Output measurement matrix of shape ``(ny, N)`` where ``ny`` is the
        number of output channels and ``N`` is the number of time samples.
        Each row represents one output channel over time.

    Attributes
    ----------
    ts : ndarray
        Time vector of shape ``(N,)``.
    us : ndarray
        Input signals of shape ``(nu, N)``.
    ys : ndarray
        Output measurements of shape ``(ny, N)``.

    Notes
    -----
    **Data Organization**:
        The class follows the convention that time varies along the second
        dimension (columns), while different signals vary along the first
        dimension (rows).

    **Validation**:
        The class automatically validates that:

        - Time vector is one-dimensional
        - Input and output matrices are two-dimensional
        - All time dimensions are consistent (same N)
        - Data types are compatible with numerical operations

    **Tree Compatibility**:
        As a `@struct`, ``Timeseries`` objects can be:

        - Flattened, reconstructed, and manipulated using tree utilities
        - Stored and manipulated as structured data
        - Used in optimization algorithms that expect tree parameters

    **Immutability**:
        Instances are frozen (immutable) after creation, preventing accidental
        modification of data during analysis. Use the ``replace`` method to
        create modified copies when needed.

    **Length and Indexing**:
        The class supports indexing and length retrieval, allowing easy
        access to specific time samples or slices of the data.

    Raises
    ------
    ValueError
        If time vector is not one-dimensional.
    ValueError
        If input or output matrices are not two-dimensional.
    ValueError
        If time dimensions are inconsistent between ts, us, and ys.

    See Also
    --------
    pem : Uses Timeseries objects for parameter estimation
    """

    ts: np.ndarray
    us: np.ndarray
    ys: np.ndarray

    def __post_init__(self):
        if self.ts.ndim != 1:
            raise ValueError("Time vector must be one-dimensional.")
        if self.ys.ndim != 2:
            raise ValueError(
                "Output measurements must be two-dimensional with shape (ny, nt)."
            )
        if self.us.ndim != 2:
            raise ValueError(
                "Input signals must be two-dimensional with shape (nu, nt)."
            )
        if self.ts.size != self.ys.shape[1]:
            raise ValueError(
                "Time vector size must match the number of time points in ys."
            )
        if self.ts.size != self.us.shape[1]:
            raise ValueError(
                "Time vector size must match the number of time points in us."
            )

    def __len__(self):
        """Return the number of time samples."""
        return self.ts.size

    def __getitem__(self, index):
        """Get a slice of the time series data."""
        return Timeseries(
            ts=self.ts[index],
            us=self.us[:, index],
            ys=self.ys[:, index],
        )



import os
import shutil
import tempfile
from typing import NamedTuple

import numpy as np
import pytest

import archimedes as arc
from archimedes._core._codegen._renderer import (
    ArduinoRenderer,
    _extract_protected_regions,
    _render_template,
)

# TODO:
# - Test data type explicit specification
# - Test data type inference
# - Test with static args
# - Test re-importing with casadi extern


@pytest.fixture
def temp_dir():
    """Create a temporary directory and file path for testing."""
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


def func(x, y):
    return x, np.sin(y) * x


@pytest.fixture()
def scalar_func():
    return func


# Create arrays of the right shape and dtype
x_type = np.array([1, 2], dtype=float)
y_type = np.array(3, dtype=float)


def compare_files(expected_file, output_dir):
    expected_output = os.path.join(
        os.path.dirname(__file__),
        f"fixtures/{expected_file}",
    )

    # Load expected output
    with open(expected_output, "r") as f:
        expected = f.read()

    # Load actual output
    with open(output_dir, "r") as f:
        actual = f.read()

    print("#### ACTUAL ####")
    print(actual)

    # Compare (normalize whitespace to handle line endings)
    assert expected.strip() == actual.strip()


def check_in_file(file, pattern):
    with open(file, "r") as f:
        content = f.read()
        assert pattern in content


class TestCodegen:
    def _check_files(self, temp_dir, func_name):
        # Check that the files were created
        assert os.path.exists(f"{temp_dir}/{func_name}_kernel.c")
        assert os.path.exists(f"{temp_dir}/{func_name}_kernel.h")
        assert os.path.exists(f"{temp_dir}/{func_name}.c")
        assert os.path.exists(f"{temp_dir}/{func_name}.h")

        # Compare to expected output
        compare_files(f"{func_name}.h", f"{temp_dir}/{func_name}.h")
        compare_files(f"{func_name}.c", f"{temp_dir}/{func_name}.c")

    def test_basic_codegen(self, temp_dir, scalar_func):
        kwargs = {
            "float_type": np.float32,
            "output_dir": temp_dir,
        }

        # Generate code for the function.  This iteration will also compile
        arc.codegen(
            scalar_func, (x_type, y_type), return_names=("x_new", "z"), **kwargs
        )
        self._check_files(temp_dir, "func")

        # Pre-compile the function for the remaining tests
        func = arc.compile(scalar_func, name="func", return_names=("x_new", "z"))

        # Check that the kernel header includes a proper function signature
        check_in_file(f"{temp_dir}/{func.name}_kernel.h", "int func")

        # Check that the API header includes correct functions
        check_in_file(f"{temp_dir}/{func.name}.h", "int func_init")
        check_in_file(f"{temp_dir}/{func.name}.h", "int func_step")

        c_file = f"{func.name}.c"
        h_file = f"{func.name}.h"

        # Run with integer arguments to check type conversion
        # Because CasADi treats everything as a float, this will get
        # converted to float in the generated code.
        y_type_int = np.array(3, dtype=int)
        arc.codegen(func, (x_type, y_type_int), **kwargs)
        check_in_file(f"{temp_dir}/{h_file}", "float y;")
        check_in_file(f"{temp_dir}/{c_file}", "arg->y = 3.000000f;")

        # Run with a specified kwarg, also test for double support
        kwargs["kwargs"] = {"y": 5.0}
        kwargs["float_type"] = np.float64
        arc.codegen(func, (x_type,), **kwargs)
        check_in_file(f"{temp_dir}/{h_file}", "double y;")
        check_in_file(f"{temp_dir}/{c_file}", "arg->y = 5.0;")

    def test_nested_codegen(self, temp_dir):
        kwargs = {
            "float_type": np.float32,
            "output_dir": temp_dir,
        }

        @arc.struct
        class EmptyStruct:
            arr: np.ndarray  # Empty array
            static_field: str = arc.field(static=True, default="static")

            @arc.struct
            class InnerStruct:
                value: float

        class Point(NamedTuple):
            x: float
            y: float

        @arc.struct
        class Cluster:
            center: Point
            points: list[Point]  # Array of structs
            weights: np.ndarray  # Simple array
            inner: EmptyStruct.InnerStruct

        def nested_func(
            scalar: float,
            arr: np.ndarray,
            clusters: list[
                Cluster
            ],  # array of structs containing arrays of named tuples
            empty_struct: EmptyStruct,
            static_str: str = "default",  # Static arg should be ignored
        ) -> float:
            return scalar + clusters[0].points[1].x

        # Pre-compile the function
        func = arc.compile(
            nested_func,
            name=nested_func.__name__,
            return_names=("z",),
            static_argnames=("static_str",),
        )

        # Initial arguments
        args = (
            42.0,  # scalar
            np.array([1.0, 2.0, 3.0]),  # arr
            [
                Cluster(
                    center=Point(1.0, 2.0),
                    points=[
                        Point(1.0, 2.0),
                        Point(2.0, 3.0),
                        Point(3.0, 4.0),
                    ],
                    weights=np.array([0.1, 0.2, 0.3]),
                    inner=EmptyStruct.InnerStruct(value=3.14),
                ),
                Cluster(
                    center=Point(4.0, 5.0),
                    points=[
                        Point(4.0, 5.0),
                        Point(5.0, 6.0),
                    ],
                    weights=np.array([0.4, 0.5, 0.6]),
                    inner=EmptyStruct.InnerStruct(value=2.71),
                ),
            ],
            EmptyStruct(arr=np.array([])),  # empty struct
        )
        # First, raise an error (second Cluster only has two points)
        with pytest.raises(
            ValueError,
            match="All items in list 'clusters' must have the same structure.",
        ):
            arc.codegen(func, args, **kwargs)

        # Fix the error and generate code
        args[2][1].points.append(Point(6.0, 7.0))  # Fix by adding a point
        arc.codegen(func, args, **kwargs, debug=True)
        self._check_files(temp_dir, func.name)

    def test_dict_codegen(self, temp_dir):
        kwargs = {
            "float_type": np.float32,
            "output_dir": temp_dir,
        }

        def dict_func(
            config: dict[str, float],  # {"lr": 0.01, "momentum": 0.9}
            bounds: tuple[float, float],  # (0.0, 1.0)
            empty_dict: dict[str, float],  # {} - empty
            empty_list: list[float],  # [] - empty
            single_tuple: tuple[float],  # (42.0,) - single element
            none_arg: None,  # empty
        ) -> dict[str, float | None]:
            return {"result": config["lr"] + bounds[0], "none_res": None}

        # Pre-compile the function
        func = arc.compile(dict_func, name=dict_func.__name__, return_names=("output",))

        args = ({"lr": 0.01, "momentum": 0.9}, (0.0, 1.0), {}, [], (42.0,), None)

        arc.codegen(func, args, **kwargs, debug=True)
        self._check_files(temp_dir, func.name)

    def test_array_codegen(self, temp_dir):
        kwargs = {
            "float_type": np.float32,
            "output_dir": temp_dir,
        }

        @arc.struct
        class EdgeCase:
            empty: np.ndarray  # []
            single: np.ndarray  # [1]

        def array_func(
            zero_d: float,  # 0D array that becomes scalar
            one_d_single: np.ndarray,  # [1] array
            one_d_normal: np.ndarray,  # [5] array
            two_d_normal: np.ndarray,  # [[1,2],[3,4]] array
            edge_case: EdgeCase,
        ) -> float:
            return np.sum(two_d_normal), zero_d + one_d_single[0], edge_case

        # Pre-compile the function
        func = arc.compile(
            array_func,
            name=array_func.__name__,
            return_names=("sum", "z", "edge_out"),
        )

        args = (
            1.0,
            np.array([1.0]),
            np.arange(5),
            np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
            EdgeCase(
                empty=np.array([]),
                single=np.array([1.0]),
            ),
        )

        arc.codegen(func, args, **kwargs, debug=True)
        self._check_files(temp_dir, func.name)

    def test_error_handling(self, scalar_func, temp_dir):
        # Test with unknown return names.  By design choice, this raises an error
        # in order to help generate more readable code.
        with pytest.raises(arc.CodegenError, match=r"Return names must be provided"):
            arc.codegen(scalar_func, (x_type, y_type))

        func = arc.compile(scalar_func, name="func", return_names=("x_new", "z"))

        # Can't use non-numeric inputs
        with pytest.raises(TypeError, match="data type 'string' not understood"):
            arc.codegen(func, (x_type, "string"))

        # Should be impossible to do this anyway, but for completeness, check that
        # the low-level codegen function raises an error if the target path is not
        # the CWD.
        specialized_func, _ = func._specialize(x_type, y_type)
        with pytest.raises(RuntimeError):
            specialized_func.codegen("invalid/func.c", {})

        # Invalid arg name (collision with `_t`)
        def invalid_arg_func(x_t):
            return x_t

        with pytest.raises(
            arc.CodegenError, match=r"Argument name 'x_t' cannot end with '_t'"
        ):
            arc.codegen(invalid_arg_func, (x_type,), return_names=("y",))

        os.remove("invalid_arg_func_kernel.c")
        os.remove("invalid_arg_func_kernel.h")


class TestExtractProtectedRegions:
    def test_basic_extraction(self):
        # Create a temporary file with test content
        content = """// Some code
    // PROTECTED-REGION-START: imports
    #include <stdlib.h>
    // PROTECTED-REGION-END
    // More code
    // PROTECTED-REGION-START: main
    printf("Hello World");
    // PROTECTED-REGION-END
    """
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp:
            temp.write(content)
            temp_path = temp.name

        try:
            # Test extraction functionality
            regions = _extract_protected_regions(temp_path)

            # Check results
            assert len(regions) == 2
            assert "imports" in regions
            assert "main" in regions
            assert "#include <stdlib.h>" in regions["imports"]
            assert 'printf("Hello World");' in regions["main"]

        finally:
            # Clean up
            os.unlink(temp_path)

    def test_nonexistent(self):
        filename = "nonexistent_file.c"
        regions = _extract_protected_regions(filename)
        assert regions == {}


@pytest.fixture
def context():
    # Basic context for consistent driver template rendering
    # Note this needs to match the test_func above
    return {
        "filename": "gen",
        "app_name": "test_app",
        "function_name": "test_func",
        "sample_rate": 0.01,
        "float_type": "float",
        "int_type": "int",
        "inputs": [
            {
                "type": "float",
                "name": "x",
                "dims": "2",
                "initial_value": "{1.0, 2.0}",
                "is_addr": False,
            },
            {
                "type": "float",
                "name": "y",
                "dims": None,
                "initial_value": "3.0",
                "is_addr": True,
            },
        ],
        "outputs": [
            {
                "type": "float",
                "name": "x_new",
                "dims": "2",
                "is_addr": False,
            },
            {
                "type": "float",
                "name": "z",
                "dims": "2",
                "is_addr": False,
            },
        ],
    }


class TestRender:
    @pytest.mark.parametrize(
        "template,expected_file",
        [
            ("c_app", "expected_c_app.c"),
            ("arduino", "expected_arduino.ino"),
        ],
    )
    def test_initial_render(self, template, expected_file, temp_dir, context):
        extension = expected_file.split(".")[-1]
        filename = f"{context['app_name']}.{extension}"
        output_path = os.path.join(temp_dir, filename)

        # Render the template
        _render_template(template, context, output_path=output_path)
        compare_files(expected_file, output_path)

    # @pytest.mark.parametrize("driver_type", ["c", "arduino"])
    # def test_default_output_path(self, driver_type, context):
    #     renderer = _render_template(driver_type, context, output_path=None)

    #     # Check that the file exists with the default name
    #     assert os.path.exists(renderer.default_output_path)
    #     os.remove(renderer.default_output_path)

    def test_invalid_renderer(self, context):
        with pytest.raises(ValueError, match=r"Template .* not found."):
            _render_template("invalid", context)

        with pytest.raises(ValueError, match=r"Template must be a .*"):
            _render_template(type(self), context)

    def test_direct_renderer(self, temp_dir, context):
        # Pass the Arduino renderer directly
        output_path = os.path.join(temp_dir, "sketch.ino")
        _render_template(ArduinoRenderer, context, output_path=output_path)

    def test_preserve_protected(self, temp_dir, context):
        output_path = os.path.join(temp_dir, f"{context['app_name']}.c")

        # Initial render
        _render_template("c_app", context, output_path=output_path)

        # Modify a protected region
        with open(output_path, "r") as f:
            content = f.read()

        # Insert custom code into a protected region
        modified_content = content.replace(
            "// PROTECTED-REGION-START: main\n",
            '// PROTECTED-REGION-START: main\n    printf("Custom code\\n");\n',
        )

        with open(output_path, "w") as f:
            f.write(modified_content)

        # Re-render with same context
        _render_template("c_app", context, output_path=output_path)

        # Verify protected region was preserved
        with open(output_path, "r") as f:
            final_content = f.read()

        assert 'printf("Custom code\\n");' in final_content

    def test_context_changes(self, temp_dir, context):
        output_path = os.path.join(temp_dir, f"{context['app_name']}.c")

        # Render with initial context
        _render_template("c_app", context, output_path=output_path)

        # Modify a protected region
        with open(output_path, "r") as f:
            content = f.read()

        modified_content = content.replace(
            "// PROTECTED-REGION-START: main\n",
            '// PROTECTED-REGION-START: main\n    printf("Custom code\\n");\n',
        )

        with open(output_path, "w") as f:
            f.write(modified_content)

        # Updated context with different function name
        context["function_name"] = "func2"

        # Re-render with new context
        _render_template("c_app", context, output_path=output_path)

        # Check results
        with open(output_path, "r") as f:
            final_content = f.read()

        # Protected region should be preserved
        assert 'printf("Custom code\\n");' in final_content

        # Function name should be updated
        assert "func2" in final_content
        assert "func1" not in final_content


if __name__ == "__main__":
    tmp_dir = "tmp"
    os.makedirs(tmp_dir, exist_ok=True)
    # TestCodegen().test_dict_codegen(tmp_dir)
    # TestCodegen().test_basic_codegen(tmp_dir, func)
    # TestCodegen().test_array_codegen(tmp_dir)
    TestCodegen().test_nested_codegen(tmp_dir)



import ast
import re

try:
    from IPython.display import Markdown, display
except ImportError:
    raise ImportError("IPython is required to use these utilities")


__all__ = ["display_text", "extract_c_function"]


def display_text(content, show=True, language="c"):
    md = Markdown(f"```{language}\n{content}\n```")

    if show:
        display(md)
    else:
        return md


def extract_c_function(filename, function_name, show=True):
    """
    Extract a specific function from a C file using regex

    Args:
        filename: Path to the C file
        function_name: Name of the function to extract

    Returns:
        The extracted function as a string, or None if not found
    """
    try:
        with open(filename, "r") as f:
            source = f.read()

        # Pattern to match function definition
        pattern = re.compile(
            r".*" + re.escape(function_name) + r"\s*\([^)]*\)\s*{", re.MULTILINE
        )

        match = pattern.search(source)
        if not match:
            print(f"Function '{function_name}' not found in '{filename}'")
            return None

        # Find the start position of the function
        start_pos = match.start()

        # Count braces to find the end of the function
        brace_count = 0
        in_function = False
        end_pos = start_pos

        for i in range(start_pos, len(source)):
            if source[i] == "{":
                in_function = True
                brace_count += 1
            elif source[i] == "}":
                brace_count -= 1
                if in_function and brace_count == 0:
                    end_pos = i + 1
                    break

        # Extract the function code
        extracted_code = source[start_pos:end_pos]

        # Display in a markdown cell
        return display_text(extracted_code, show=show, language="c")

    except FileNotFoundError:
        print(f"File '{filename}' not found")
        return None


def extract_py_function(
    filename: str,
    function_name: str,
    show: bool = True,
    include_decorators: bool = True,
) -> str | None:
    """
    Extract a specific function from a Python file using AST parsing

    Args:
        filename: Path to the Python file
        function_name: Name of the function to extract
        show: Whether to display the result
        include_decorators: Whether to include decorators in the extraction

    Returns:
        The extracted function as a string, or None if not found
    """
    return _extract_py_definition(
        filename, function_name, "function", show, include_decorators
    )


def extract_py_class(
    filename: str, class_name: str, show: bool = True, include_decorators: bool = True
) -> str | None:
    """
    Extract a specific class from a Python file using AST parsing

    Args:
        filename: Path to the Python file
        class_name: Name of the class to extract
        show: Whether to display the result
        include_decorators: Whether to include decorators in the extraction

    Returns:
        The extracted class as a string, or None if not found
    """
    return _extract_py_definition(
        filename, class_name, "class", show, include_decorators
    )


def _extract_py_definition(
    filename: str,
    name: str,
    definition_type: str = "function",
    show: bool = True,
    include_decorators: bool = True,
) -> str | None:
    """
    Extract a function or class from a Python file using AST parsing

    Args:
        filename: Path to the Python file
        name: Name of the function/class to extract
        definition_type: Either "function" or "class"
        show: Whether to display the result
        include_decorators: Whether to include decorators in the extraction

    Returns:
        The extracted code as a string, or None if not found
    """
    try:
        with open(filename, "r", encoding="utf-8") as f:
            source = f.read()

        tree = ast.parse(source)
        lines = source.splitlines()

        target_node_type = (
            ast.FunctionDef if definition_type == "function" else ast.ClassDef
        )

        for node in ast.walk(tree):
            if isinstance(node, target_node_type) and node.name == name:
                start_line = node.lineno - 1  # AST uses 1-based indexing

                # Include decorators if requested
                if include_decorators and node.decorator_list:
                    decorator_start = min(d.lineno for d in node.decorator_list) - 1
                    start_line = decorator_start

                # Try to get end line (available in Python 3.8+)
                if hasattr(node, "end_lineno") and node.end_lineno is not None:
                    end_line = node.end_lineno
                else:
                    # Error for older Python versions
                    raise SyntaxError(
                        "Python version must be 3.8 or higher to extract definitions "
                        "without end_lineno."
                    )

                extracted_code = "\n".join(lines[start_line:end_line])
                return display_text(extracted_code, show=show, language="python")

        print(f"{definition_type.capitalize()} '{name}' not found in '{filename}'")
        return None

    except FileNotFoundError:
        print(f"File '{filename}' not found")
        return None
    except SyntaxError as e:
        print(f"Syntax error in '{filename}': {e}")
        return None



import io
import sys

import numpy as np
from scipy import signal

import archimedes as arc


def test_callback():
    def f(x, y):
        print(f"f called: {x=}, {y=}")
        return x * (y + 3)

    @arc.compile
    def call_f(x, y):
        result_shape_dtypes = x  # Output has same type as input
        return arc.callback(f, result_shape_dtypes, x, y)

    x, y = np.array([1.0, 2.0]), 3.0
    z = call_f(x, y)
    assert np.allclose(z, f(x, y))
    assert z.shape == x.shape


def test_unsupported_op():
    def calc_sum_psd(x):
        # Computation that is not supported symbolically
        f, Pxx = signal.welch(x)  # noqa: N806
        print(f)
        return sum(Pxx[f > 0.1])

    # Call from within a compiled function
    @arc.compile
    def sum_psd(x):
        result_shape_dtypes = 0.0  # Template data type for
        return arc.callback(calc_sum_psd, result_shape_dtypes, x)

    x = np.arange(1024)

    y = calc_sum_psd(x)
    assert np.allclose(y, sum_psd(x))


def test_tree_structured_callback():
    """Test with tree-structured data."""

    def tree_func(data):
        return {"doubled": data["values"] * 2, "squared": data["values"] ** 2}

    @arc.compile
    def tree_model(data):
        result_shape_dtypes = {"doubled": data["values"], "squared": data["values"]}
        return arc.callback(tree_func, result_shape_dtypes, data)

    data = {"values": np.array([1.0, 2.0, 3.0])}
    result = tree_model(data)
    expected = tree_func(data)

    assert all(np.allclose(result[k], expected[k]) for k in expected)


def test_multiple_arguments():
    """Test with multiple input arguments."""

    def multi_arg_func(x, y, z):
        return x * y + z

    @arc.compile
    def multi_model(x, y, z):
        result_shape_dtypes = x
        return arc.callback(multi_arg_func, result_shape_dtypes, x, y, z)

    x = np.array([1.0, 2.0])
    y = 3.0
    z = np.array([0.5, 1.0])

    result = multi_model(x, y, z)
    expected = multi_arg_func(x, y, z)

    assert np.allclose(result, expected)


def test_error_propagation():
    """Test that errors in the callback function are properly propagated."""

    def error_func(x):
        if x[0] < 0:
            raise ValueError("Negative input not allowed")
        return np.sqrt(x)

    try:

        @arc.compile
        def error_model(x):
            result_shape_dtypes = x  # Output has same type as input
            return arc.callback(error_func, result_shape_dtypes, x)

        # This should fail at runtime
        error_model(np.array([-1.0, 4.0]))

        # If we get here, something went wrong
        assert False, "Should have raised an exception"

    # TODO: Ideally this would return the actual ValueError, but it's not
    # straightforward to extract the exception from CasADi, so the RuntimeError
    # is the best we can do for now.
    except RuntimeError as e:
        assert "Negative input not allowed" in str(e)
        print("Error propagation test passed")


def test_debug_print():
    def print_func(x):
        print("Value: {}".format(x))
        return x * 2

    # Create a function that uses the print functionality
    @arc.compile
    def print_test(x):
        return arc.callback(print_func, x, x)

    # Redirect stdout
    captured_output = io.StringIO()
    sys.stdout = captured_output

    # Call the function with test data
    x = np.array([1.0, 2.0, 3.0])
    result = print_test(x)

    # Restore stdout
    sys.stdout = sys.__stdout__

    # Check that the output contains the expected text
    assert "Value: [1. 2. 3.]" in captured_output.getvalue()

    # Also verify that the function returns the correct result
    assert np.allclose(result, x * 2)



"""Interface for solving ordinary differential equations.

This module has two main functions: `integrator` and `odeint`.  The `integrator`
function is a transformation that creates a "forward map" for the given function.
This forward map is a new function that integrates the ODE defined by the original
function.  The `odeint` function is a simple wrapper around the `integrator` function
that calls the generated forward map with the given initial state and time span.
`odeint` has a similar interface to `scipy.integrate.solve_ivp`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Sequence, cast

import casadi as cs

from archimedes import tree
from archimedes._core import (
    FunctionCache,
    SymbolicArray,
    _as_casadi_array,
    sym,
    sym_like,
)

__all__ = ["integrator", "odeint"]

if TYPE_CHECKING:
    from archimedes.typing import ArrayLike, NDArray


def integrator(
    func: Callable,
    method: str = "cvodes",
    atol: float = 1e-6,
    rtol: float = 1e-3,
    static_argnames: str | Sequence[str] = (),
    t_eval: NDArray | None = None,
    name: str | None = None,
    **options,
) -> FunctionCache:
    """Create an ODE solver function from a dynamics function.

    Transforms a function defining an ODE into a function that propagates the ODE
    system forward in time. The resulting function can be used repeatedly with
    different initial conditions and parameters, and supports automatic differentiation
    through the solution.

    Parameters
    ----------
    func : callable
        The function defining the ODE dynamics. Must have the signature
        ``func(t, x, *args) -> x_t``, where ``t`` is the current time,
        ``x`` is the current state, and ``x_t`` is the derivative of
        the state with respect to time.
    method : str, optional
        Integration method to use. Default is ``"cvodes"``, which uses the
        SUNDIALS CVODES solver (suitable for both stiff and non-stiff systems).
        See CasADi documentation for other available methods.
    atol : float, optional
        Absolute tolerance for the solver. Default is ``1e-6``.
    rtol : float, optional
        Relative tolerance for the solver. Default is ``1e-3``.
    static_argnames : tuple of str, optional
        Names of arguments to ``func`` that should be treated as static
        (not traced symbolically).
    t_eval : array_like, optional
        Times at which to evaluate the solution. If provided, the integrator
        will return states at these times. If None, the integrator will
        only return the final state.
    name : str, optional
        Name for the created integrator function. If None, a name is
        automatically generated based on the dynamics function's name.
    **options : dict
        Additional options to pass to the CasADi integrator.

    Returns
    -------
    callable
        If ``t_eval`` is None, a function with signature
        ``solver(x0, t_span, *args) -> xf``, where ``x0`` is the initial state,
        ``t_span`` is a tuple ``(t0, tf)`` giving the integration time span, and ``xf``
        is the final state.

        If ``t_eval`` is provided, a function with signature
        ``solver(x0, *args) -> xs``, where ``xs`` is an array of states at the times in
        ``t_eval``.

    Notes
    -----
    When to use this function:

    - For repeated ODE solves with different initial conditions or parameters
    - When embedding ODE solutions in optimization problems or larger models
    - When sensitivity analysis requires derivatives of the solution
    - For improved performance over loop-based ODE integration

    The ODE solver is specified by the ``method`` argument. The default is ``"cvodes"``,
    which uses the SUNDIALS CVODES solver and is suitable for both stiff and non-stiff
    systems. This is the recommended choice for most applications.

    See the [CasADi](https://web.casadi.org/python-api/#integrator) documentation for a
    complete list of available configuration options to pass to the `options` argument.

    Conceptual model:
    This function constructs a computational graph representation of the entire
    ODE solve operation. The resulting function is more efficient than repeatedly
    calling :py:func:``odeint`` because:

    - It pre-compiles the solver for a specific state dimension and parameter structure
    - It leverages CasADi's compiled C++ implementation for high-performance integration
    - It enables automatic differentiation through the entire solution

    Limitations:

    - Only scalar and vector states are supported (shapes like ``(n,)`` or ``(n,1)``)
    - If ``t_eval`` is provided, the function caches the evaluation times internally,
      meaning new evaluation times require recompiling the function

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Example 1: Simple pendulum
    >>> def pendulum(t, x):
    >>>     # Simple pendulum dynamics.
    >>>     theta, omega = x
    >>>     return np.array([omega, -9.81 * np.sin(theta)], like=x)
    >>>
    >>> # Create an integrator
    >>> solver = arc.integrator(pendulum)
    >>>
    >>> # Solve for multiple initial conditions
    >>> x0_1 = np.array([0.1, 0.0])  # Small angle
    >>> x0_2 = np.array([1.5, 0.0])  # Large angle
    >>> t_span = (0.0, 10.0)
    >>>
    >>> x1_final = solver(x0_1, t_span)
    >>> x2_final = solver(x0_2, t_span)
    >>>
    >>> # Example 2: Parameterized system
    >>> def lotka_volterra(t, x, params):
    >>>     # Lotka-Volterra predator-prey model with parameters.
    >>>     a, b, c, d = params
    >>>     prey, predator = x
    >>>     return np.array([
    >>>         a * prey - b * prey * predator,      # prey growth rate
    >>>         c * prey * predator - d * predator   # predator growth rate
    >>>     ], like=x)
    >>>
    >>> # Create an integrator with evaluation times
    >>> t_eval = np.linspace(0, 20, 100)
    >>> solver = arc.integrator(lotka_volterra, t_eval=t_eval, method="cvodes")
    >>>
    >>> # Solve with different parameters
    >>> x0 = np.array([10.0, 5.0])  # Initial population
    >>> params1 = np.array([1.5, 0.3, 0.2, 0.8])
    >>> params2 = np.array([1.2, 0.2, 0.3, 1.0])
    >>>
    >>> solution1 = solver(x0, params1)  # Shape: (2, 100)
    >>> solution2 = solver(x0, params2)
    >>>
    >>> # Example 3: Differentiating through ODE solutions
    >>> target = solution1
    >>> @arc.compile(kind="MX")
    >>> def cost(x0, params):
    >>>     # Cost function based on final state after integration.
    >>>     final_state = solver(x0, params)
    >>>     return np.sum((final_state - target)**2)
    >>>
    >>> # Compute gradient of cost with respect to parameters
    >>> dJ_dp = arc.grad(cost, argnums=1)
    >>> print(dJ_dp(x0, params1))
    [0. 0. 0. 0.]
    >>> print(dJ_dp(x0, params2))
    [-11757.72435691 -32985.03384074  43301.00963694 -27677.60203032]

    See Also
    --------
    odeint : One-time ODE solution for a specific initial value problem
    """

    if not isinstance(func, FunctionCache):
        func = FunctionCache(func, static_argnames=static_argnames)

    options = {
        **options,
        "abstol": atol,
        "reltol": rtol,
    }

    # Function to determine the shape of the output of the ODE solver
    def _shape_inference(x0):
        # If `t_eval` is None, then the ODE solution will be a single state vector.
        if t_eval is None:
            return x0.shape

        # Otherwise the ODE solution will be an array of states at the evaluation
        # times. If `x0.shape == (n, 1)` then the output shape will be
        # `(n, len(t_eval))`, flattening over the empty dimension.
        return (*x0.shape[:1], len(t_eval))

    # Define a function that will integrate the ODE through the time span.
    # This function will be evaluated with SymbolicArray objects.
    def _forward_map(x0, t_span, *args) -> ArrayLike:
        if len(x0.shape) > 1 and x0.shape[1] > 1:
            raise ValueError(
                f"Only scalar and vector states are supported. Got shape {x0.shape}"
            )

        ret_shape = _shape_inference(x0)
        ret_dtype = x0.dtype

        # We have to flatten all of the symbolic user-defined arguments
        # into a single vector to pass to CasADi.  If there is static data,
        # this needs to be stripped out and passed separately.
        if static_argnames or func.static_argnums:
            # The first two arguments are (t, x), so skip these in checking
            # for static arguments.
            static_argnums = [i - 2 for i in func.static_argnums]
            static_args, sym_args, _arg_types = func.split_args(static_argnums, *args)

        else:
            # No static data - all arguments can be treated symbolically
            static_argnums = []
            sym_args = args

        # We cannot use a concatenation of various symbolic variables as an input to
        # the integrator, so we need to create a fresh symbolic variable for this
        p_orig, unravel = tree.ravel(sym_args)
        p = sym("p", kind="MX", shape=p_orig.shape, dtype=p_orig.dtype)

        # Create a new set of "args" that includes the fresh symbolic parameters
        # for use in evaluating the ODE function symbolically.
        args: list[ArrayLike] = list(unravel(p))  # type: ignore[no-redef]
        # If the function has static arguments, we need to interleave them back in the
        # original order.
        for i in static_argnums:
            args.insert(i, static_args[i])  # type: ignore[attr-defined]

        # Define consistent time and state variables
        t, x = sym("t", kind="MX"), sym_like(x0, name="x0", kind="MX")
        xdot = func(t, x, *args)

        # For type checking
        p = cast(SymbolicArray, p)
        xdot = cast(SymbolicArray, xdot)

        ode = {
            "t": t._sym,
            "x": x._sym,
            "ode": xdot._sym,
        }

        if p.size != 0:
            ode["p"] = p._sym

        solver = cs.integrator("solver", method, ode, *t_span, options)

        # Before calling the CasADi solver interface, make sure everything is
        # either a CasADi symbol or a NumPy array
        p_arg = False if p_orig is None else p_orig
        x0, p_arg = map(_as_casadi_array, (x0, p_arg))  # type: ignore[assignment]

        # The return is a dictionary with the final state of the system
        # and other information.  We will only return the final state.
        sol = solver(x0=x0, p=p_arg)

        return SymbolicArray(
            sol["xf"],
            dtype=ret_dtype,
            shape=ret_shape,
        )

    if t_eval is None:
        # The first two names to the function are time and state,
        # otherwise they will be user-defined
        arg_names = ("x0", "t_span", *func.arg_names[2:])
        static_argnames = ("t_span", *static_argnames)
        _wrapped_forward_map = _forward_map

    else:
        # The CasADi interface expects `t0` to be a scalar, but allows
        # `tf` to be an array of evaluation times.  Here we will close
        # over the evaluation times because this cannot be used as a
        # hashable key for the compiled function cache (i.e. a static arg).
        t0, tf = t_eval[0], t_eval

        def _wrapped_forward_map(x0, *args) -> ArrayLike:  # type: ignore
            return _forward_map(x0, (t0, tf), *args)

        arg_names = ("x0", *func.arg_names[2:])

    if name is None:
        name = f"{func.name}_odeint"

    _wrapped_forward_map.__name__ = name

    return FunctionCache(
        _wrapped_forward_map,
        arg_names=arg_names,
        static_argnames=static_argnames,
        kind="MX",
    )


def odeint(
    func: Callable,
    t_span: tuple[float, float],
    x0: ArrayLike,
    method: str = "cvodes",
    t_eval: NDArray | None = None,
    atol: float = 1e-6,
    rtol: float = 1e-3,
    args: Sequence[Any] | None = None,
    static_argnames: str | Sequence[str] = (),
    **options,
) -> ArrayLike:
    """Integrate a system of ordinary differential equations.

    Solves the initial value problem defined by the ODE system:

    .. math::
        \\dot{x} = f(t, x, \\theta)

    from ``t=t_span[0]`` to ``t=t_span[1]`` with initial conditions ``x0``.

    Parameters
    ----------
    func : callable
        Right-hand side of the ODE system. The calling signature is
        ``func(t, x, *args)``, where ``t`` is a scalar and ``x`` is an ndarray with
        shape ``(n,)``. func must return an array with shape ``(n,)``.
    t_span : tuple of float
        Interval of integration ``(t0, tf)``.
    x0 : array_like
        Initial state.
    method : str, optional
        Integration method to use. Default is ``"cvodes"``, which uses the
        SUNDIALS CVODES solver (suitable for both stiff and non-stiff systems).
    t_eval : array_like or None, optional
        Times at which to store the computed solution. If ``None``, the solver
        only returns the final state.
    atol : float, optional
        Absolute tolerance for the solver. Default is ``1e-6``.
    rtol : float, optional
        Relative tolerance for the solver. Default is ``1e-3``.
    args : tuple, optional
        Additional arguments to pass to the ODE function.
    static_argnames : tuple of str, optional
        Names of arguments to `func` that should be treated as static
        (not traced symbolically).
    **options : dict
        Additional options to pass to the CasADi integrator.

    Returns
    -------
    array_like
        If ``t_eval`` is None, the state at the final time ``t_span[1]``.
        If ``t_eval`` is provided, an array containing the solution evaluated at
        ``t_eval`` with shape ``(n, len(t_eval))``.

    Notes
    -----

    The underlying SUNDIALS solvers use adaptive step size control to balance
    accuracy and computational efficiency.  However, unlike SciPy's ``solve_ivp``
    function and many other ODE solver interfaces, it is not possible to output
    the solution at the times actually used by the solver (or to create a "dense"
    output using an interpolant). This is because of the requirement that all input
    and output arrays be a fixed size in order to construct the symbolic graph. The
    solution at output times specified by ``t_eval`` is computed by interpolating the
    solution at the times used by the solver.

    The ODE solver is specified by the ``method`` argument. The default is ``"cvodes"``,
    which uses the SUNDIALS CVODES solver and is suitable for both stiff and non-stiff
    systems. This is the recommended choice for most applications.

    See the `CasADi <https://web.casadi.org/python-api/#integrator/>`_ documentation
    for a complete list of available configuration options to pass to the ``options``
    argument.

    Conceptual model:
    This function is a wrapper around the :py:func:``integrator`` function that creates
    and immediately calls an ODE solver for the given initial conditions and
    parameters. It provides a simple interface similar to SciPy's ``solve_ivp``
    but leverages CasADi's efficient symbolic implementation and interface to
    SUNDIALS solvers.

    Unlike :py:func:`integrator`, which creates a reusable solver function,
    ``odeint`` performs a single solve operation. If you need to solve the same ODE
    system repeatedly with different initial conditions or parameters, it's more
    efficient to create an ``integrator`` function once and reuse it.

    Limitations:

    - Only scalar and vector states are supported (shapes like ``(n,)`` or ``(n,1)``)
    - For tree-structured states, currently the user must define a wrapper function
      to flatten and reconstruct the tree structure

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>> import matplotlib.pyplot as plt
    >>>
    >>> # Example 1: Simple harmonic oscillator
    >>> def f(t, y):
    >>>     return np.array([y[1], -y[0]], like=y)  # y'' = -y
    >>>
    >>> t_span = (0, 10)
    >>> y0 = np.array([1.0, 0.0])  # Initial displacement and velocity
    >>> ts = np.linspace(*t_span, 100)
    >>>
    >>> # Solve ODE
    >>> ys = arc.odeint(f, t_span, y0, t_eval=ts)
    >>>
    >>> plt.figure()
    >>> plt.plot(ts, ys[0, :], label='position')
    >>> plt.plot(ts, ys[1, :], label='velocity')
    >>> plt.legend()
    >>> plt.show()
    >>>
    >>> # Example 2: Parametrized dynamics
    >>> def pendulum(t, y, g, L):
    >>>     theta, omega = y
    >>>     return np.array([omega, -(g/L) * np.sin(theta)])
    >>>
    >>> y0 = np.array([np.pi/6, 0.0])  # 30 degrees initial angle
    >>> t_span = (0, 5)
    >>> ys = arc.odeint(pendulum, t_span, y0, args=(9.81, 1.0))

    See Also
    --------
    integrator : Create a reusable ODE solver function
    scipy.integrate.solve_ivp : SciPy's ODE solver interface
    """

    solver = integrator(
        func,
        method=method,
        atol=atol,
        rtol=rtol,
        t_eval=t_eval,
        static_argnames=static_argnames,
        **options,
    )
    if args is None:
        args = ()

    if t_eval is not None:
        xs: ArrayLike = solver(x0, *args)

    else:
        xs: ArrayLike = solver(x0, t_span, *args)  # type: ignore[no-redef]

    return xs



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Archimedes"
copyright = "2025, Pine Tree Labs, LLC"
author = "Jared Callaham"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # "myst_parser",
    "myst_nb",
    "sphinx_design",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinxcontrib.googleanalytics",
]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "attrs_block",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

nb_kernel_rgx_aliases = {
    ".*": "archimedes"  # Use the registered Jupyter kernel name
}
nb_execution_mode = "cache"
nb_execution_timeout = 60
nb_execution_cache_path = ".jupyter_cache"
nb_execution_allow_errors = False  # Fail build on exceptions
nb_execution_raise_on_error = True  # Raise immediately on error
nb_execution_excludepatterns = [
    "**/*.ipynb",  # Exclude all notebooks by default
    "experimental/*",  # Skip WIP content
    "benchmarks/*",  # Skip long-running benchmarks
]

autosummary_generate = True
autosummary_imported_members = True

# Maximum signature line length before breaking into multiple lines
maximum_signature_line_length = 88
autodoc_preserve_defaults = True
# autodoc_typehints = "description"  # or "signature", "both", "none"

templates_path = ["_templates"]
exclude_patterns = []

googleanalytics_id = "G-DMLVH3TEDW"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# html_theme = "alabaster"
html_theme = "furo"
html_theme_options = {
    # Light mode variables
    "light_css_variables": {
        "color-brand-primary": "#D35400",  # copper orange
        "color-brand-content": "#C0392B",  # ember red
        "color-admonition-background": "rgba(211, 84, 0, 0.1)",  # transparent orange
        "color-background-primary": "#F5F5F5",  # light gray
        "color-background-secondary": "#EEEEEE",  # slightly darker gray
        "color-foreground-primary": "#2A2A2A",  # dark charcoal
        "color-foreground-secondary": "#5D4037",  # rich brown
        "color-link": "#C0392B",  # ember red
        "color-link-hover": "#D35400",  # copper orange
    },
    # Dark mode variables
    "dark_css_variables": {
        "color-brand-primary": "#F1C40F",  # pale gold
        "color-brand-content": "#D35400",  # copper orange
        "color-admonition-background": "rgba(241, 196, 15, 0.1)",  # transparent gold
        "color-background-primary": "#2A2A2A",  # dark charcoal
        "color-background-secondary": "#1A1A1A",  # darker charcoal
        "color-foreground-primary": "#F5F5F5",  # light gray
        "color-foreground-secondary": "#DDDDDD",  # slightly darker light gray
        "color-link": "#D35400",  # copper orange
        "color-link-hover": "#F1C40F",  # pale gold
    },
    "top_of_page_buttons": [],
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/pinetreelabs/archimedes",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}

html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "custom.css",
]

html_js_files = []

# Add favicon configuration
html_favicon = "_static/favicon.ico"



"""Integration test of OCPs with analytic solution"""

# ruff: noqa: N802
# ruff: noqa: N803
# ruff: noqa: N806

import numpy as np

from archimedes.experimental import coco as cc


class TestDoubleIntegrator:
    def minimum_control(self, t, x, u, p):
        return u**2

    def minimum_time(self, t, x, u, p):
        return 1.0

    def double_integrator(self, t, x, u, p):
        return np.array([x[1], u[0]], like=x)

    def lagrange_cost(self, x0, t0, xf, tf, q, p):
        return q

    def _make_min_ctrl_problem(self, x0, xf, t0, tf, N=6):
        boundary_conditions = [
            cc.start_time(t0),
            cc.end_time(tf),
            cc.initial_condition(x0),
            cc.final_condition(xf),
        ]

        # Add bounds to test path constraints
        path_constraints = [
            cc.state_bounds([-10.0, -10.0], [10.0, 10.0]),
            cc.control_bounds([-10.0], [10.0]),
        ]

        # Define the optimal control problem
        ocp = cc.OptimalControlProblem(
            nx=2,
            nu=1,
            ode=self.double_integrator,
            quad=self.minimum_control,
            cost=self.lagrange_cost,
            boundary_constraints=boundary_conditions,
            path_constraints=path_constraints,
        )

        # Discretize the domain
        domain = cc.RadauFiniteElements(N=[N], knots=[])
        return ocp, domain

    def test_min_ctrl(self, plot=False):
        x0, xf = np.array([0.0, 0.0]), np.array([1.0, 0.0])
        t0, tf = 0.0, 1.0

        ocp, domain = self._make_min_ctrl_problem(x0, xf, t0, tf, N=6)

        # Linearly interpolate initial guess
        def x_guess(t):
            return x0 + (t - t0) * (xf - x0) / (tf - t0)

        # Solve the optimal control problem with IPOPT
        sol = ocp.solve(domain, t_guess=(t0, tf), x_guess=x_guess)

        #
        # Check against analytic solution
        #
        def x_ex(t):
            return 3 * t**2 - 2 * t**3

        def u_ex(t):
            return 6 - 12 * t

        t0, tf = sol.t0, sol.tf
        t_plt = np.linspace(t0, tf, 10)
        x_plt = sol.x(t_plt)
        u_plt = sol.u(t_plt)

        assert np.allclose(x_plt[:, 0], x_ex(t_plt), atol=1e-3)
        assert np.allclose(u_plt.squeeze(), u_ex(t_plt), atol=1e-3)

        if plot:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(2, 1, figsize=(7, 3), sharex=True)
            ax[0].scatter(sol.tp, sol.xp[:, 0], c="tab:red", label="Optimal trajectory")
            ax[0].plot(t_plt, x_plt[:, 0], c="tab:red")
            ax[0].plot(t_plt, x_ex(t_plt), "k--", lw=2, label="Exact solution")
            ax[0].legend()
            ax[0].grid()
            ax[0].set_ylabel(r"$x$")
            ax[1].scatter(sol.tp[:-1], sol.up, c="tab:red")
            ax[1].plot(t_plt, u_plt, c="tab:red")
            ax[1].plot(t_plt, u_ex(t_plt), "k--", lw=2)
            ax[1].grid()
            ax[1].set_ylabel(r"$u$")
            ax[1].set_xlabel(r"$t$")

            plt.show()

    def test_default_initial_guess(self):
        x0, xf = np.array([0.0, 0.0]), np.array([1.0, 0.0])
        t0, tf = 0.0, 2.0

        ocp, domain = self._make_min_ctrl_problem(x0, xf, t0, tf, N=4)

        initial_guess, _lower_bound, _upper_bound = ocp.initialize(domain)
        x, u, t0, tf, p = ocp.unpack_dvs(initial_guess, domain)
        assert np.allclose(x, 0.0)
        assert np.allclose(u, 0.0)
        assert np.allclose(t0, 0.0)
        assert np.allclose(tf, 1.0)

    def test_min_time(self):
        x0, xf = np.array([0.0, 0.0]), np.array([1.0, 0.0])
        u_min, u_max = -10.0, 10.0
        t0 = 0.0
        N = 4

        boundary_conditions = [
            cc.start_time(t0),
            cc.initial_condition(x0),
            cc.final_condition(xf),
        ]

        # Add bounds to test path constraints
        path_constraints = [
            cc.control_bounds([u_min], [u_max]),
        ]

        # Define the optimal control problem
        ocp = cc.OptimalControlProblem(
            nx=2,
            nu=1,
            ode=self.double_integrator,
            quad=self.minimum_time,
            cost=self.lagrange_cost,
            boundary_constraints=boundary_conditions,
            path_constraints=path_constraints,
        )

        # Discretize the domain.  We can split into two segments to test "bang-bang"
        # control, knowing that the optimal switching time is halfway through based
        # on the symmetry of the problem.
        domain = cc.RadauFiniteElements(N=[N, N], knots=[0.0])

        # Linearly interpolate initial guess
        tf_guess = 1.0

        def x_guess(t):
            return x0 + (t - t0) * (xf - x0) / (tf_guess - t0)

        # Solve the optimal control problem with IPOPT
        sol = ocp.solve(domain, t_guess=(t0, tf_guess), x_guess=x_guess)

        assert np.allclose(sol.tf, 0.632455549)

        # Check for "bang-bang" control
        t1 = sol.tp[sol.tp < sol.tf / 2]
        t2 = sol.tp[sol.tp >= sol.tf / 2]

        assert np.allclose(sol.u(t1), u_max)
        assert np.allclose(sol.u(t2), u_min)


class TestProjectile:
    def projectile_dynamics(self, t, x, u, p):
        g0 = 9.81

        vx, vy = x[2], x[3]
        ax = np.zeros_like(vx)
        ay = np.zeros_like(vy) - g0
        return np.array([vx, vy, ax, ay], like=x)

    def dummy_quad(self, t, x, u, p):
        return 0.0

    def max_range(self, x0, t0, xf, tf, q, p):
        return -xf[0]

    def test_projectile(self):
        """What angle should you launch a projectile for maximum range?

        Tests a problem with a discrete parameter (launch angle)
        """
        v0 = 1.0

        # Parametric initial condition
        def initial_condition(boundary_data):
            x0 = boundary_data.x0
            theta = boundary_data.p[0]
            return x0 - np.array(
                [0.0, 0.0, v0 * np.cos(theta), v0 * np.sin(theta)], like=x0
            )

        def final_condition(boundary_data):
            return boundary_data.xf[1]

        t0, tf_guess = 0.0, 2.0
        boundary_conditions = [
            cc.start_time(t0),
            cc.Constraint(initial_condition, nc=4),
            cc.Constraint(final_condition, nc=1),
            cc.parameter_bounds([0.0], [np.pi]),
        ]

        # Define the optimal control problem
        ocp = cc.OptimalControlProblem(
            nx=4,
            nu=1,
            np=1,
            ode=self.projectile_dynamics,
            quad=self.dummy_quad,
            cost=self.max_range,
            boundary_constraints=boundary_conditions,
        )

        domain = cc.RadauFiniteElements(N=[3], knots=[])
        sol = ocp.solve(domain, t_guess=(t0, tf_guess))

        # Analytic solution to the projectile problem
        theta_ex = np.pi / 4
        tf_ex = 2 * v0 * np.sin(theta_ex) / 9.81
        range_ex = np.cos(theta_ex) * tf_ex
        x0_ex = np.array([0.0, 0.0, np.cos(theta_ex), np.sin(theta_ex)])
        xf_ex = np.array([range_ex, 0.0, np.cos(theta_ex), -np.sin(theta_ex)])

        assert np.allclose(sol.p[0], theta_ex)
        assert np.allclose(sol.t0, 0.0)
        assert np.allclose(sol.tf, tf_ex, atol=1e-3)
        assert np.allclose(sol.xp[0], x0_ex)
        assert np.allclose(sol.xp[-1], xf_ex, atol=1e-3)


class TestHohmannTransfer:
    def test_single_stage(self, plot=False):
        """Classical ideal Hohmann transfer"""
        mu = 1.0
        r0 = 1.0
        rf = 1.5
        t0 = 0.0

        def j(t, x, phi, p):
            return 0.0

        def f(t, x, phi, p):
            r, θ, u, v = x
            return np.array(
                [
                    u,
                    v / r,
                    v**2 / r - mu / r**2,
                    -u * v / r,
                ],
                like=x,
            )

        def cost(x0, t0, xf, tf, q, p):
            return sum(p**2)  # Total delta-v expenditure

        # Add impulsive (tangential) velocity at the beginning
        def initial_constraint(boundary_data):
            dv1 = boundary_data.p[0]
            x0_transfer = np.array(
                [r0, 0.0, 0.0, np.sqrt(mu / r0) + dv1], like=boundary_data.x0
            )
            return x0_transfer - boundary_data.x0

        # Final state of the transfer orbit should be the target orbit,
        # less the impulsive (tangential) velocity change. The final anomaly
        # is not constrained
        def terminal_constraint(boundary_data):
            dv2 = boundary_data.p[1]
            xf = boundary_data.xf[[0, 2, 3]]
            xf_circular = np.array(
                [rf, 0.0, np.sqrt(mu / rf) - dv2], like=boundary_data.xf
            )
            return xf_circular - xf

        # Discretize the domain
        N = [20]
        knots = []
        domain = cc.RadauFiniteElements(N=N, knots=knots)

        # Boundary conditions
        bcs = [
            cc.start_time(t0),
            cc.Constraint(initial_constraint, 4),
            cc.Constraint(terminal_constraint, 3),
        ]

        # Define the problem in three stages
        ocp = cc.OptimalControlProblem(
            nx=4,
            nu=0,
            np=2,
            ode=f,
            quad=j,
            cost=cost,
            boundary_constraints=bcs,
        )

        # Linearly interpolate initial guess
        x0_guess = np.array([r0, 0.0, 0.0, np.sqrt(mu / r0)])
        xf_guess = np.array([rf, np.pi, 0.0, np.sqrt(mu / rf)])
        tf_guess = 4.5

        def x_guess(t):
            return x0_guess + (t - t0) * (xf_guess - x0_guess) / (tf_guess - t0)

        sol = ocp.solve(domain, t_guess=(t0, tf_guess), x_guess=x_guess)

        # Hohmann solution
        dv1_ex = np.sqrt(mu / r0) * (np.sqrt(2 * rf / (r0 + rf)) - 1)
        dv2_ex = np.sqrt(mu / rf) * (1 - np.sqrt(2 * r0 / (r0 + rf)))

        assert np.allclose(sol.p, np.array([dv1_ex, dv2_ex]), atol=1e-3)
        assert np.allclose(sol.xp[-1, 1], np.pi, atol=1e-3)  # Anomaly of 180 degrees

        if plot:
            import matplotlib.pyplot as plt

            t_plt = np.linspace(t0, sol.tf, 100)
            x_plt = sol.x(t_plt)

            fig, ax = plt.subplots(1, 1, figsize=(4, 4))

            r, θ = x_plt[:, 0], x_plt[:, 1]
            x, y = r * np.cos(θ), r * np.sin(θ)

            rf = r[-1]
            r0 = r[0]
            rmax = max(r)

            ax.plot(x, y, c="xkcd:brick")

            θ = np.linspace(0, 2 * np.pi, 1000)
            ax.plot(r0 * np.cos(θ), r0 * np.sin(θ), c="grey", ls="--")
            ax.plot(rf * np.cos(θ), rf * np.sin(θ), c="k", ls="--")

            ax.set_xlim([-1.2 * rmax, 1.2 * rmax])
            ax.set_ylim([-1.2 * rmax, 1.2 * rmax])
            ax.grid()
            plt.show()

    def test_multistage(self, plot=False):
        """Test multistage problems by adding a specified final anomaly"""
        mu = 1.0
        r0 = 1.0
        rf = 1.5

        nx, nu = 4, 0
        x0 = np.array([r0, 0.0, 0.0, np.sqrt(mu / r0)])
        xf = np.array([rf, 2 * np.pi, 0.0, np.sqrt(mu / rf)])
        t0 = 0.0

        def j(t, x, phi, p):
            return 0.0

        def f(t, x, phi, p):
            r, θ, u, v = x
            return np.array(
                [
                    u,
                    v / r,
                    v**2 / r - mu / r**2,
                    -u * v / r,
                ],
                like=x,
            )

        def cost_stage1(x0, t0, xf, tf, q, p):
            return sum(p**2)  # Total delta-v expenditure

        def cost_stage2(x0, t0, xf, tf, q, p):
            return sum(p**2)  # Total delta-v expenditure

        # State continuity between stages
        def stage_constraint(boundary_data):
            bd1, bd2 = boundary_data
            bd2_x0 = bd1.xf
            bd2_x0[3] += bd1.p  # Add delta-v to tangential velocity
            return bd2_x0 - bd2.x0

        def terminal_constraint(boundary_data):
            dv2 = boundary_data.p
            stage_xf = boundary_data.xf
            xf_circular = np.array([xf[0], xf[1], xf[2], xf[3] - dv2], like=stage_xf)
            return xf_circular - stage_xf

        # Discretize the domain (same for all three stages)
        N = [20]
        knots = []
        stage_domain = cc.RadauFiniteElements(N=N, knots=knots)

        # Boundary conditions for each stage
        bcs1 = [cc.start_time(t0), cc.initial_condition(x0)]
        bcs2 = [cc.Constraint(terminal_constraint, 4)]

        # Define the problem in three stages
        stage1 = cc.OptimalControlProblem(
            nx=nx,
            nu=nu,
            np=1,
            ode=f,
            quad=j,
            cost=cost_stage1,
            boundary_constraints=bcs1,
        )

        stage2 = cc.OptimalControlProblem(
            nx=nx,
            nu=nu,
            np=1,
            ode=f,
            quad=j,
            cost=cost_stage2,
            boundary_constraints=bcs2,
        )

        # Define the full problem
        ocp = cc.MultiStageOptimalControlProblem(
            stages=[stage1, stage2],
            stage_constraints=[
                cc.Constraint(stage_constraint, 4),
            ],
        )

        domain = [stage_domain, stage_domain]

        # Linearly interpolate initial guess
        tf_guess = 5.0

        def x_guess(t):
            return x0 + (t - t0) * (xf - x0) / (tf_guess - t0)

        sol = ocp.solve(domain, x_guess=x_guess)

        # Hohmann solution
        dv1_ex = np.sqrt(mu / r0) * (np.sqrt(2 * rf / (r0 + rf)) - 1)
        dv2_ex = np.sqrt(mu / rf) * (1 - np.sqrt(2 * r0 / (r0 + rf)))

        assert np.allclose(sol[0].p, dv1_ex, atol=1e-3)
        assert np.allclose(sol[1].p, dv2_ex, atol=1e-3)
        assert np.allclose(sol[0].xp[-1, 1], np.pi)  # Anomaly of 180 degrees
        assert np.allclose(sol[0].xp[-1, :2], sol[1].xp[0, :2])  # State continuity
        assert np.allclose(sol[1].xp[-1, 1], 2 * np.pi)  # Anomaly of 360 degrees

        if plot:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(1, 1, figsize=(4, 4))

            for i in range(len(sol)):
                t_plt = np.linspace(sol[i].t0, sol[i].tf, 100)
                x_plt = sol[i].x(t_plt)
                r, θ = x_plt[:, 0], x_plt[:, 1]
                x, y = r * np.cos(θ), r * np.sin(θ)

                ax.plot(x, y, lw=2, c="xkcd:brick")

            θ = np.linspace(0, 2 * np.pi, 1000)
            ax.plot(r0 * np.cos(θ), r0 * np.sin(θ), c="grey", ls="--")
            ax.plot(rf * np.cos(θ), rf * np.sin(θ), c="k", ls="--")

            ax.set_xlim([-1.2 * rf, 1.2 * rf])
            ax.set_ylim([-1.2 * rf, 1.2 * rf])
            ax.grid()
            plt.show()


if __name__ == "__main__":
    TestDoubleIntegrator().test_min_ctrl(plot=True)



import numpy as np
import pytest

from archimedes._core import FunctionCache, SymbolicArray, compile, sym


def f(x):
    return x
    return x**2 + 1


class TestCompile:
    def test_construction(self):
        f_sym = compile(f)
        assert isinstance(f_sym, FunctionCache)
        assert f_sym.name == "f"
        assert f_sym.arg_names == ["x"]
        assert f_sym._compiled == {}

        # Test recompilation
        f_sym2 = compile(f_sym)
        assert f_sym2 is f_sym

    def test_naming(self):
        # Lambda function
        f_sym = compile(lambda x: x)
        assert f_sym.name.startswith("lambda")

        # Callable class
        class _Functor:
            def __call__(self, x):
                return 2 * x

        func = _Functor()
        assert not hasattr(func, "__name__")
        f_sym = compile(func)
        assert f_sym.name.startswith("function")

    def test_eval_numeric(self):
        f_sym = compile(f)

        # Scalar
        x = 2
        assert f_sym(x) == f(x)

        # NumPy Arrays of different shapes (will have to "recompile")
        for x in (
            np.array([1, 2, 3]),
            np.array([[1, 2], [3, 4]]),
        ):
            y = f_sym(x)
            print(y)
            assert isinstance(y, np.ndarray)
            assert np.all(y == f(x))
            assert y.shape == x.shape

    def test_eval_symbolic(self):
        f_sym = compile(f)

        x = sym("x", shape=(), dtype=np.int32)
        y = f_sym(x)
        assert isinstance(y, SymbolicArray)
        assert y.shape == ()
        assert y.dtype == np.int32

    def test_multiple_returns(self):
        # Test a function that returns multiple values
        def f2(x):
            return x**2, x + 1

        f_sym = compile(f2)

        x = np.array(2.0, dtype=np.float64)
        y1, y2 = f_sym(x)
        y1_expected, y2_expected = f2(x)
        assert y1 == y1_expected
        assert y2 == y2_expected

        assert isinstance(y1, np.ndarray)
        assert isinstance(y2, np.ndarray)
        assert y1.dtype == x.dtype
        assert y2.dtype == x.dtype

    def test_tree_returns(self):
        @compile
        def g(x, y):
            return (x, y), 2 * x

        out = g(0.0, 1.0)
        print(f"{out=}")

    def test_static(self):
        # Test compiling function with static argument specified by number

        def f2(a, x):
            return a * x

        # Create functions specifying static data both by number and by name
        for f_sym in (
            compile(f2, static_argnums=0),
            compile(f2, static_argnames="a"),
        ):
            # Nothing compiled yet
            assert len(f_sym._compiled) == 0

            a = 2.0
            x = np.array(3.0, dtype=np.float64)
            y = f_sym(a, x)
            assert y == a * x
            # Only one function compiled so far
            assert len(f_sym._compiled) == 1

            # Call again with different static data
            a = 3.0
            y = f_sym(a, x)
            assert y == a * x
            # Should have compiled a new function
            assert len(f_sym._compiled) == 2

    def test_kwargs(self):
        a0 = 2.0

        def f(x, a=a0):
            return a * x

        f_sym = compile(f)
        x = 3.0

        # Call without specifying `a`
        y = f_sym(x)
        assert y == a0 * x

        # Specify `a` by keyword
        a = 3.0
        y = f_sym(x, a=a)
        assert y == a * x

        # Specify `a` by position
        y = f_sym(x, a)
        assert y == a * x

        # Specify both args by keyword
        y = f_sym(x=x, a=a0)
        assert y == a0 * x

    def test_numeric_returns(self):
        # Should be able to return both numeric and symbolic arrays
        @compile
        def f(x):
            return x, np.array([3.0, 4.0]), True, 3

        x = np.array(2.0, dtype=np.float64)
        y1, y2, y3, y4 = f(x)

        assert all(isinstance(y, np.ndarray) for y in (y1, y2, y3, y4))

        assert y1 == x
        assert np.all(y2 == np.array([3.0, 4.0]))
        assert y3.dtype == np.bool_
        assert y3
        assert y4.dtype == int
        assert y4 == 3

    def test_error_handling(self):
        # Test function with varargs
        def f(x, *args):
            return x + sum(args)

        with pytest.raises(ValueError):
            compile(f)

        # Specify both static_argnums and static_argnames
        def f(a, x):
            return a * x

        with pytest.raises(ValueError):
            compile(f, static_argnums=(0,), static_argnames=("a",))

        # Static arg not in signature
        with pytest.raises(ValueError):
            compile(f, static_argnames=("b",))

        # Call with incorrect number of arguments (raised by `inspect`)
        f = compile(f, static_argnames=("a",))
        with pytest.raises(TypeError):
            f(2.0)

        # Return something besides an array
        def f(x):
            return "abc"

        with pytest.raises(TypeError):
            compile(f)(0.0)

        # Return variable number of values
        def f(x, flag):
            if flag:
                return 2 * x
            else:
                return 2 * x, 3 * x

        f = compile(f, static_argnames=("flag",))
        f(2.0, True)  # OK the first time
        with pytest.raises(ValueError):
            f(2.0, False)  # Fail the second time

        # Unhashable static arg (list)
        def f(a, x):
            return x

        f = compile(f, static_argnames=("a",))
        with pytest.raises(ValueError):
            f([1.0], 2.0)

    def test_modified_array(self):
        # Check that in-place modification of an array doesn't change
        # the original value (because dummy args are created during compilation)
        x = np.zeros((2, 3))

        @compile
        def f(x):
            x[0] = x[0] + 1.0
            return x

        y = f(x)
        assert np.allclose(x, 0.0)
        assert np.allclose(y[0], 1.0)
        assert np.allclose(y[1], 0.0)



# This code modifies code from JAX

# Copyright (c) 2021 The JAX Authors
# Licensed under Apache License 2.0
# https://github.com/jax-ml/jax

# Modifications and additions to the original code:
# Copyright (c) 2025 Pine Tree Labs, LLC
# Licensed under the GNU General Public License v3.0

# SPDX-FileCopyrightText: 2021 The JAX Authors
# SPDX-FileCopyrightText: 2025 Pine Tree Labs, LLC
# SPDX-License-Identifier: GPL-3.0-or-later

# As a combined work, use of this code requires compliance with the GNU GPL v3.0.
# The original license terms are included below for attribution:

# === Apache License 2.0 ===
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools as it
from functools import partial, reduce
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Hashable,
    Iterable,
    Iterator,
    NamedTuple,
    TypeVar,
    cast,
)

import numpy as np

from ._registry import _registry, unzip2

if TYPE_CHECKING:
    from ..typing import ArrayLike, Tree

    T = TypeVar("T", bound=Tree)
    V = TypeVar("V")


class TreeDef(NamedTuple):
    node_data: None | tuple[type, Hashable]
    children: tuple[TreeDef, ...]
    num_leaves: int

    def unflatten(self, xs: list[Any]) -> Any:
        return tree_unflatten(self, xs)

    @property
    def tree_str(self) -> str:
        stars = ["*"] * self.num_leaves
        return cast(str, self.unflatten(stars))

    def __repr__(self) -> str:
        return (f"TreeDef({self.tree_str})").replace("'*'", "*")

    def __eq__(self, other: object) -> bool:
        other = cast(TreeDef, other)
        return (
            self.node_data == other.node_data
            and self.children == other.children
            and self.num_leaves == other.num_leaves
        )


LEAF = TreeDef(None, (), 1)
NONE_DEF = TreeDef(None, (), 0)


#
# Flatten/unflatten functions
#
def tree_flatten(
    x: Tree, is_leaf: Callable[[Any], bool] | None = None
) -> tuple[list[ArrayLike], TreeDef]:
    """
    Flatten a tree into a list of leaves and a treedef.

    This function recursively traverses the tree and extracts all leaf values
    while recording the structure. This is useful when you need to apply operations
    to all leaf values or convert structured data to a flat representation.

    Parameters
    ----------
    x : Tree
        A tree to be flattened. Here, a tree is a nested structure of containers
        (lists, tuples, dicts, etc) and leaves (arrays, scalars, objects not registered
        as trees).
    is_leaf : callable, optional
        A function that takes a tree as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types are used.

    Returns
    -------
    leaves : list
        A list of all leaf values from the tree.
    treedef : TreeDef
        A structure definition that can be used to reconstruct the original
        tree using `unflatten`.

    Notes
    -----
    In this context, a tree is defined as a nested structure of:

    - Containers: recognized container types like lists, tuples, and dictionaries
    - Leaves: arrays, scalars, or custom objects not recognized as containers

    When to use:

    - When you need to extract all leaf values from a nested structure
    - When you need to convert between structured and flat representations

    When converting to/from a flat vector it will typically be more convenient to
    use :py:func:`ravel` instead of this function.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Simple tree with nested containers
    >>> data = {"a": np.array([1.0, 2.0]), "b": {"c": np.array([3.0])}}
    >>>
    >>> # Flatten the tree
    >>> leaves, treedef = arc.tree.flatten(data)
    >>> print(leaves)
    [array([1., 2.]), array([3.])]
    >>>
    >>> # Use treedef to reconstruct the tree
    >>> reconstructed = arc.tree.unflatten(treedef, leaves)
    >>> print(reconstructed)
    {'a': array([1., 2.]), 'b': {'c': array([3.])}}

    See Also
    --------
    unflatten : Reconstruct a tree from leaves and a treedef
    leaves : Extract just the leaf values from a tree
    structure : Extract just the structure from a tree
    ravel : Flatten a tree into a single 1D array
    """
    children_iter, treedef = _tree_flatten(x, is_leaf)
    return list(children_iter), treedef


def _tree_flatten(
    x: Tree, is_leaf: Callable[[Any], bool] | None
) -> tuple[Iterable, TreeDef]:
    if x is None:
        return [], NONE_DEF

    _tree_flatten_leaf = partial(_tree_flatten, is_leaf=is_leaf)

    node_type = type(x)
    # If the node is a namedtuple, use the tuple flatten/unflatten functions
    if isinstance(x, tuple) and hasattr(x, "_fields"):
        node_type = tuple
    if node_type not in _registry or (is_leaf is not None and is_leaf(x)):
        return [x], LEAF

    children, node_metadata = _registry[node_type].to_iter(x)
    children_flat, child_trees = unzip2(map(_tree_flatten_leaf, children))
    flattened = list(it.chain.from_iterable(children_flat))

    node_data = (type(x), node_metadata)
    treedef = TreeDef(
        node_data=node_data,
        children=tuple(child_trees),
        num_leaves=len(flattened),
    )
    return flattened, treedef


def tree_unflatten(treedef: TreeDef, xs: list[ArrayLike]) -> Tree:
    """
    Reconstruct a tree from a list of leaves and a treedef.

    This function is the inverse of :py:func:`flatten`. It takes a list of leaf values
    and a tree definition, and reconstructs the original tree structure.

    Parameters
    ----------
    treedef : TreeDef
        A tree definition, typically produced by :py:func:`flatten` or
        :py:func:`structure`.
    xs : list[ArrayLike]
        A list of leaf values to be placed in the reconstructed tree.
        The length must match the number of leaves in ``treedef``.

    Returns
    -------
    tree : Tree
        The reconstructed tree with the same structure as defined by treedef
        and with leaf values from ``xs``.

    Notes
    -----
    When converting to/from a flat vector it will typically be more convenient to
    use :py:func:`ravel` instead of this function.

    Raises
    ------
    ValueError
        If the number of leaves in ``xs`` doesn't match the expected number in
        ``treedef``.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Original tree
    >>> data = {"positions": np.array([1.0, 2.0]), "velocities": np.array([3.0, 4.0])}
    >>>
    >>> # Flatten the tree
    >>> leaves, treedef = arc.tree.flatten(data)
    >>>
    >>> # Transform the leaves (e.g., multiply by 2)
    >>> new_leaves = [leaf * 2 for leaf in leaves]
    >>>
    >>> # Reconstruct the tree with the new leaves
    >>> new_data = arc.tree.unflatten(treedef, new_leaves)
    >>> print(new_data)
    {'positions': array([2., 4.]), 'velocities': array([6., 8.])}

    See Also
    --------
    flatten : Flatten a tree into a list of leaves and a treedef
    structure : Extract just the structure from a tree
    ravel : Flatten a tree into a single 1D array
    """
    return _tree_unflatten(treedef, iter(xs))


def _tree_unflatten(treedef: TreeDef, xs: Iterator) -> Tree:
    if treedef is NONE_DEF:
        return None  # Special case for None
    if treedef.node_data is None:
        return next(xs)
        # The following was the original code, but test coverage
        # shows its's not used and it doesn't seem to be necessary
        # after introducing NONE_DEF.
        # try:
        #     return next(xs)
        # except StopIteration:
        #     return None
    else:
        children = tuple(_tree_unflatten(t, xs) for t in treedef.children)
        node_type, node_metadata = treedef.node_data

        # Special logic for NamedTuple classes
        if issubclass(node_type, tuple) and hasattr(node_type, "_fields"):
            return node_type(*children)

        return _registry[node_type].from_iter(node_metadata, children)


#
# Other utility functions
#


def tree_structure(tree: Tree, is_leaf: Callable[[Any], bool] | None = None) -> TreeDef:
    """
    Extract the structure of a tree without the leaf values.

    This function returns a :py:class:TreeDef that describes the structure of the
    tree, which can be used with :py:func:`unflatten` to reconstruct a tree with
    new leaf values.

    Parameters
    ----------
    tree : Tree
        A tree whose structure is to be determined.
    is_leaf : callable, optional
        A function that takes a tree node as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types (arrays and scalars) are used.

    Returns
    -------
    treedef : :py:class:`TreeDef`
        A tree definition that describes the structure of the input tree.

    Notes
    -----
    When to use:

    - When you need to extract just the structure of a tree for later use
    - When you want to create a template structure that can be filled with
      different leaf values
    - When you need to compare the structures of two trees

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a structured state
    >>> state = {"pos": np.array([0.0, 1.0]), "vel": np.array([2.0, 3.0])}
    >>>
    >>> # Extract the structure
    >>> treedef = arc.tree.structure(state)
    >>> print(treedef)
    TreeDef({'pos': *, 'vel': *})
    >>>
    >>> # Create a new state with the same structure but different values
    >>> zeros = [np.zeros_like(leaf) for leaf in arc.tree.leaves(state)]
    >>> initial_state = arc.tree.unflatten(treedef, zeros)
    >>> print(initial_state)
    {'pos': array([0., 0.]), 'vel': array([0., 0.])}

    See Also
    --------
    flatten : Flatten a tree into a list of leaves and a treedef
    unflatten : Reconstruct a tree from leaves and a treedef
    ravel : Flatten a tree into a single 1D array
    """
    flat, treedef = tree_flatten(tree, is_leaf)
    return treedef


def tree_leaves(
    tree: Tree, is_leaf: Callable[[Any], bool] | None = None
) -> list[ArrayLike]:
    """
    Extract all leaf values from a tree.

    This function traverses the tree and returns a list of all leaf values
    without the structure information.

    Parameters
    ----------
    tree : Tree
        A tree from which to extract leaves. Here, a tree is a nested structure of
        containers (lists, tuples, dicts, etc) and leaves (arrays, scalars, objects
        not registered as trees).
    is_leaf : callable, optional
        A function that takes a tree node as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types (arrays and scalars) are used.

    Returns
    -------
    leaves : list
        A list of all leaf values from the tree.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a structured data object
    >>> data = {"params": {"w": np.array([1.0, 2.0]), "b": 0.5},
    ...         "state": np.array([3.0, 4.0])}
    >>>
    >>> # Extract all leaf values
    >>> leaf_values = arc.tree.leaves(data)
    >>> print(leaf_values)
    [0.5, array([1., 2.]), array([3., 4.])]

    See Also
    --------
    flatten : Flatten a tree into a list of leaves and a treedef
    ravel : Flatten a tree into a single 1D array
    map : Apply a function to each leaf in a tree
    """
    flat, treedef = tree_flatten(tree, is_leaf)
    return flat


def tree_all(tree: Tree, is_leaf: Callable[[Any], bool] | None = None) -> bool:
    """
    Check if all leaves in the tree evaluate to True.

    This function traverses the tree and checks if all leaf values are truthy,
    similar to Python's built-in :py:func:`all` function but operating on all leaves
    of a tree.

    Parameters
    ----------
    tree : Tree
        A tree to check. A tree is a nested structure of containers
        (lists, tuples, dicts) and leaves (arrays, scalars, objects
        not registered as trees).
    is_leaf : callable, optional
        A function that takes a tree node as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types (arrays and scalars) are used.

    Returns
    -------
    result : bool
        True if all leaves in the tree evaluate to True, False otherwise.

    Notes
    -----
    When to use:

    - To check if all values in a structured object meet a certain condition\
      (after applying :py:func:`map` with a condition function)
    - To validate that all components of a model or state are initialized
    - As a convenient way to check properties across nested structures

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Check if all values are positive
    >>> data = {"a": np.array([1.0, 2.0]), "b": {"c": np.array([3.0])}}
    >>> is_positive = arc.tree.map(lambda x: x > 0, data)
    >>> all_positive = arc.tree.all(is_positive)
    >>> print(all_positive)
    True
    >>>
    >>> # Check if all values are greater than 1.5
    >>> is_greater = arc.tree.map(lambda x: x > 1.5, data)
    >>> all_greater = arc.tree.all(is_greater)
    >>> print(all_greater)
    False

    See Also
    --------
    map : Apply a function to each leaf in a tree
    leaves : Extract just the leaf values from a tree
    """
    flat, treedef = tree_flatten(tree, is_leaf)
    return np.all(map(np.all, flat))  # type: ignore


def tree_map(
    f: Callable,
    tree: T,
    *rest: tuple[T, ...],
    is_leaf: Callable[[Any], bool] | None = None,
) -> T:
    """
    Apply a function to each leaf in a tree.

    This function traverses the tree and applies the function ``f`` to each leaf,
    returning a new tree with the same structure but transformed leaf values.
    If additional trees are provided, the function is applied to corresponding
    leaves from all trees.

    Parameters
    ----------
    f : callable
        A function to apply to each leaf. When multiple trees are provided,
        this function should accept as many arguments as there are trees.
    tree : Any
        The main tree whose structure will be followed.
    *rest : Any
        Additional trees with exactly the same structure as the first tree.
    is_leaf : callable, optional
        A function that takes a tree node as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types (arrays and scalars) are used.

    Returns
    -------
    mapped_tree : Any
        A new tree with the same structure as ``tree`` but with leaf values
        transformed by function ``f``.

    Notes
    -----
    When to use:

    - To transform data in a structured object without changing its structure
    - To perform element-wise operations on corresponding elements of multiple trees
    - As an alternative to manually looping through nested structures
    - To apply the same operation to all arrays in a complex model

    Raises
    ------
    ValueError
        If additional trees do not have exactly the same structure as the main tree.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Single tree example
    >>> state = {"pos": np.array([1.0, 2.0]), "vel": np.array([3.0, 4.0])}
    >>>
    >>> # Double all values
    >>> doubled = arc.tree.map(lambda x: x * 2, state)
    >>> print(doubled)
    {'pos': array([2., 4.]), 'vel': array([6., 8.])}
    >>>
    >>> # Multiple trees example
    >>> state1 = {"pos": np.array([1.0, 2.0]), "vel": np.array([3.0, 4.0])}
    >>> state2 = {"pos": np.array([5.0, 6.0]), "vel": np.array([7.0, 8.0])}
    >>>
    >>> # Add corresponding leaves
    >>> combined = arc.tree.map(lambda x, y: x + y, state1, state2)
    >>> print(combined)
    {'pos': array([6., 8.]), 'vel': array([10., 12.])}

    See Also
    --------
    flatten : Flatten a tree into a list of leaves and a treedef
    leaves : Extract just the leaf values from a tree
    """
    leaves, treedef = tree_flatten(tree, is_leaf)
    flat = [leaves]
    for r in rest:
        r_flat, r_treedef = tree_flatten(r, is_leaf)
        if treedef != r_treedef:
            raise ValueError(
                "Trees must have the same structure but got treedefs: "
                f"{treedef} and {r_treedef}"
            )
        flat.append(r_flat)

    flat = [f(*args) for args in zip(*flat)]
    tree_out: T = tree_unflatten(treedef, flat)  # type: ignore
    return tree_out


def tree_reduce(
    function: Callable[[V, ArrayLike], V],
    tree: Tree,
    initializer: V,
    is_leaf: Callable[[Any], bool] | None = None,
) -> V:
    """
    Reduce a tree to a single value using a function and initializer.

    This function traverses the tree, applying the reduction function to
    each leaf and an accumulator, similar to Python's built-in :py:func:`reduce`
    but operating on all leaves of a tree.

    Parameters
    ----------
    function : callable
        A function of two arguments: (accumulated_result, leaf_value) that
        returns a new accumulated result. The function should be commutative
        and associative to ensure results are independent of traversal order.
    tree : Tree
        A tree to reduce. Here, a tree is a nested structure of containers
        (lists, tuples, dicts) and leaves (arrays or scalars).
    initializer : Any
        The initial value for the accumulator.
    is_leaf : callable, optional
        A function that takes a tree node as input and returns a boolean
        indicating whether it should be considered a leaf. If not provided,
        the default leaf types (arrays and scalars) are used.

    Returns
    -------
    result : Any
        The final accumulated value after applying the function to all leaves.

    Notes
    -----
    When to use:

    - To compute aggregate values (sum, product, etc.) across all leaf values
    - To collect statistics from a structured model
    - To implement custom reduction operations on complex data structures

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Sum all values in a nested structure
    >>> data = {"a": np.array([1.0, 2.0]), "b": {"c": np.array([3.0])}}
    >>>
    >>> # Compute the sum
    >>> def sum_leaf(acc, leaf):
    ...     return acc + sum(leaf)
    >>>
    >>> total = arc.tree.reduce(sum_leaf, data, 0.0)
    >>> print(total)
    6.0
    >>>
    >>> # Find the maximum value
    >>> def max_leaf(acc, leaf):
    ...     return np.fmax(acc, np.max(leaf))
    >>>
    >>> maximum = arc.tree.reduce(max_leaf, data, -np.inf)
    >>> print(maximum)
    3.0

    See Also
    --------
    map : Apply a function to each leaf in a tree
    leaves : Extract just the leaf values from a tree
    """
    flat, _treedef = tree_flatten(tree, is_leaf)
    return reduce(function, flat, initializer)


def is_leaf(x: Any) -> bool:
    """Check if a value is a leaf in a tree.

    Returns True if the value is not a container (i.e. is an array,
    a scalar, or None).

    Parameters
    ----------
    x : Any
        The value to check.

    Returns
    -------
    bool
        True if the value is a leaf, False otherwise.
    """
    treedef = tree_structure(x)
    return treedef is LEAF or treedef is NONE_DEF



# This code modifies code from NumPy.

# Copyright (c) 2005-2024, NumPy Developers.
# Licensed under NumPy license
# https://numpy.org/doc/stable/license.html

# Modifications and additions to the original code:
# Copyright (c) 2025 Pine Tree Labs, LLC
# Licensed under the GNU General Public License v3.0

# As a combined work, use of this code requires compliance with the GNU GPL v3.0.
# The original license terms are included below for attribution:

# === NumPy License ===

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#     * Neither the name of the NumPy Developers nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import casadi as cs
import numpy as np

from .._array_impl import DEFAULT_FLOAT, SymbolicArray, _as_casadi_array
from .._type_inference import shape_inference, type_inference
from ._array_ops import array, binary_op, unary_op


def _dot(x1, x2):
    # If either x1 or x2 are scalars, then use element-wise multiplication.
    x1_is_scalar = np.isscalar(x1) or len(x1.shape) == 0
    x2_is_scalar = np.isscalar(x2) or len(x2.shape) == 0
    if x1_is_scalar or x2_is_scalar:
        return x1 * x2

    # Otherwise, we need to use CasADi's `dot` function for the symbolic
    # component. `dot` should close over the inner dimension of x1 and the outer
    # dimension of x2. The shape inference also checks that the shapes are compatible.
    dtype = type_inference("default", x1, x2)
    x1, x2 = map(array, (x1, x2))
    shape = shape_inference("matmul", x1, x2)

    # Now that type inference is done we can discard the SymbolicArray wrapper and just
    # use the underlying data (SX or NumPy array)
    arg1, arg2 = map(_as_casadi_array, (x1, x2))

    # CasADi's `dot` function requires that the two arrays have the same shape, and
    # acts as a flattened vector dot product. However, for NumPy it is equivalent to
    # matrix multiplication.  Hence, we should only use `cs.dot` if both are column
    # vectors.
    if len(x1.shape) == 1 and len(x2.shape) == 1:
        result = cs.dot(arg1, arg2)

    else:
        # If the shape of the array is (n,) then the CasADi symbolic array will be
        # (n, 1), whereas CasADi will expect a (1, n) array for a column vector.
        # Note that initialization of the SymbolicArray enforces that if x.shape = (n,)
        # then x._sym.shape = (n, 1), so the second argument does not need to be checked
        # here.
        if len(x1.shape) == 1:
            # First argument is a SymbolicArray: transpose the underlying CasADi array
            if isinstance(x1, SymbolicArray) and arg1.shape[1] == 1:
                arg1 = arg1.T
            # First argument is an ndarray: make into a row vector
            elif isinstance(x1, np.ndarray):
                arg1 = arg1.reshape((1, -1))

        result = cs.mtimes(arg1, arg2)

    return SymbolicArray(result, dtype=dtype, shape=shape)


def _matmul(x1, x2):
    # If either x1 or x2 are scalars, then use element-wise multiplication.
    x1_is_scalar = np.isscalar(x1) or len(x1.shape) == 0
    if x1_is_scalar:
        raise ValueError("Matmul input 0 does not have enough dimensions (requires 1)")
    x2_is_scalar = np.isscalar(x2) or len(x2.shape) == 0
    if x2_is_scalar:
        raise ValueError("Matmul input 1 does not have enough dimensions (requires 1)")
    # NumPy matmul is the same as dot for 2D arrays
    return _dot(x1, x2)


@binary_op
def _floor_divide(x1, x2):
    return cs.floor(x1 / x2)


def _divmod(x1, x2, dtype=None):
    y1 = np.floor_divide(x1, x2, dtype=dtype)
    y2 = np.remainder(x1, x2, dtype=dtype)
    return y1, y2


def _hypot(x, y):
    return np.sqrt(x * x + y * y)


def _xor(a, b):
    return cs.logic_or(
        cs.logic_and(a, cs.logic_not(b)), cs.logic_and(cs.logic_not(a), b)
    )


def _radians(x):
    return x * (np.pi / 180.0)


def _degrees(x):
    return x * (180.0 / np.pi)


# List from numpy.testing.overrides.get_overridable_numpy_ufuncs()
SUPPORTED_UFUNCS = {
    "floor": unary_op(cs.floor),
    # Note that np.maximum propagates NaN
    "maximum": binary_op(cs.fmax, result_type=DEFAULT_FLOAT),
    "floor_divide": _floor_divide,
    # Note that np.minimum propagates NaN
    "minimum": binary_op(cs.fmin, result_type=DEFAULT_FLOAT),
    "fmax": binary_op(cs.fmax, result_type=DEFAULT_FLOAT),
    "modf": NotImplemented,
    "fmin": binary_op(cs.fmin, result_type=DEFAULT_FLOAT),
    "multiply": binary_op(cs.times),
    "fmod": binary_op(cs.fmod),
    "negative": unary_op(lambda x: -x),
    "frexp": NotImplemented,
    "nextafter": NotImplemented,
    "gcd": NotImplemented,
    "not_equal": binary_op(cs.ne, result_type=bool),
    "cosh": unary_op(cs.cosh, result_type=DEFAULT_FLOAT),
    "greater": binary_op(cs.gt, result_type=bool),
    "positive": NotImplemented,
    "cos": unary_op(cs.cos, result_type=DEFAULT_FLOAT),
    "greater_equal": binary_op(cs.ge, result_type=bool),
    "power": binary_op(cs.power),
    "copysign": NotImplemented,
    "heaviside": NotImplemented,
    "rad2deg": _degrees,
    "conjugate": NotImplemented,
    "hypot": _hypot,
    "radians": _radians,
    "clip": NotImplemented,
    "invert": unary_op(cs.logic_not, result_type=bool),
    "reciprocal": NotImplemented,
    "ceil": unary_op(cs.ceil),
    "isfinite": NotImplemented,
    "remainder": binary_op(cs.remainder),  # np.remainder and np.mod,
    "cbrt": NotImplemented,
    "isinf": NotImplemented,
    "right_shift": NotImplemented,
    "bitwise_xor": NotImplemented,
    "isnan": NotImplemented,
    "rint": NotImplemented,
    "bitwise_or": NotImplemented,
    "isnat": NotImplemented,
    "sign": unary_op(cs.sign),
    "_ones_like": NotImplemented,
    # Note that np.absolute handles complex
    "absolute": unary_op(cs.fabs),
    "bitwise_and": NotImplemented,
    "lcm": NotImplemented,
    "signbit": NotImplemented,
    "add": binary_op(cs.plus),
    "arctanh": unary_op(cs.atanh, result_type=DEFAULT_FLOAT),
    "ldexp": NotImplemented,
    "sin": unary_op(cs.sin, result_type=DEFAULT_FLOAT),
    "arccos": unary_op(cs.acos, result_type=DEFAULT_FLOAT),
    "arctan2": binary_op(cs.atan2, result_type=DEFAULT_FLOAT),
    "left_shift": NotImplemented,
    "sinh": unary_op(cs.sinh, result_type=DEFAULT_FLOAT),
    "arccosh": unary_op(cs.acosh, result_type=DEFAULT_FLOAT),
    "arctan": unary_op(cs.atan, result_type=DEFAULT_FLOAT),
    "less": binary_op(cs.lt, result_type=bool),
    "spacing": NotImplemented,
    "arcsin": unary_op(cs.asin, result_type=DEFAULT_FLOAT),
    "less_equal": binary_op(cs.le, result_type=bool),
    "sqrt": unary_op(cs.sqrt, result_type=DEFAULT_FLOAT),
    "arcsinh": unary_op(cs.asinh, result_type=DEFAULT_FLOAT),
    "deg2rad": _radians,
    "log": unary_op(cs.log, result_type=DEFAULT_FLOAT),
    "square": NotImplemented,
    "degrees": _degrees,
    "log10": unary_op(cs.log10, result_type=DEFAULT_FLOAT),
    "subtract": binary_op(cs.minus),
    "divide": binary_op(cs.rdivide, result_type=DEFAULT_FLOAT),
    "true_divide": binary_op(cs.rdivide, result_type=DEFAULT_FLOAT),
    "log1p": unary_op(cs.log1p, result_type=DEFAULT_FLOAT),
    "tan": unary_op(cs.tan, result_type=DEFAULT_FLOAT),
    "divmod": _divmod,
    "log2": NotImplemented,
    "tanh": unary_op(cs.tanh, result_type=DEFAULT_FLOAT),
    "equal": binary_op(cs.eq, result_type=bool),
    "logaddexp": NotImplemented,
    "trunc": NotImplemented,
    "exp": unary_op(cs.exp, result_type=DEFAULT_FLOAT),
    "logaddexp2": NotImplemented,
    "exp2": NotImplemented,
    "logical_and": binary_op(cs.logic_and, result_type=bool),
    "expm1": NotImplemented,
    "logical_not": unary_op(cs.logic_not, result_type=bool),
    "fabs": unary_op(cs.fabs),
    "logical_or": binary_op(cs.logic_or, result_type=bool),
    "float_power": NotImplemented,
    "logical_xor": binary_op(_xor, result_type=bool),
    "matmul": _matmul,
}
